"""Util functions."""

from collections.abc import Callable
import contextlib
from datetime import datetime, timedelta, timezone
import functools
import logging
import re
import subprocess
from typing import TYPE_CHECKING, Any, Literal, Optional

import pandas as pd
import pendulum
import pyodbc
import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import (
    ConnectionError,
    HTTPError,
    ReadTimeout,
    Timeout,
)
from requests.packages.urllib3.util.retry import Retry
from urllib3.exceptions import ProtocolError

from viadot.exceptions import APIError, ValidationError
from viadot.orchestration.prefect.utils import DynamicDateHandler
from viadot.signals import SKIP


if TYPE_CHECKING:
    from viadot.sources.base import Source


with contextlib.suppress(ImportError):
    import pyspark.sql.dataframe as spark


def slugify(name: str) -> str:
    """Slugify a string."""
    return name.replace(" ", "_").lower()


def handle_api_request(
    url: str,
    auth: tuple | None = None,
    params: dict[str, Any] | None = None,
    headers: dict[str, Any] | None = None,
    timeout: tuple = (3.05, 60 * 30),
    method: Literal["GET", "POST", "DELETE"] = "GET",
    data: str | None = None,
    verify: bool = True,
) -> requests.Response:
    """Send an HTTP request to the specified URL using the provided parameters.

    Args:
        url (str): The URL to send the request to.
        auth (tuple, optional): A tuple of (username, password) for basic
            authentication. Defaults to None.
        params (Dict[str, Any], optional): A dictionary of query string parameters.
            Defaults to None.
        headers (Dict[str, Any], optional): A dictionary of HTTP headers to include with
            the request. Defaults to None.
        timeout (tuple, optional): A tuple of (connect_timeout, read_timeout) in
            seconds. Defaults to (3.05, 60 * 30).
        method (Literal["GET", "POST", "DELETE"], optional): The HTTP method to use for
            the request. Defaults to "GET".
        data (str, optional): The request body data as a string. Defaults to None.
        verify (bool, optional): Whether to verify certificates. Defaults to True.

    Returns:
        requests.Response: The HTTP response object.
    """
    session = requests.Session()
    retry_strategy = Retry(
        total=3,
        status_forcelist=[429, 500, 502, 503, 504],
        backoff_factor=1,
        allowed_methods=["GET", "POST", "DELETE"],
    )
    adapter = HTTPAdapter(max_retries=retry_strategy)

    session.mount("http://", adapter)
    session.mount("https://", adapter)

    return session.request(
        method=method,
        url=url,
        auth=auth,
        params=params,
        headers=headers,
        timeout=timeout,
        data=data,
        verify=verify,
    )


def _handle_response(
    response: requests.Response, timeout: tuple = (3.05, 60 * 30)
) -> requests.Response:
    url = response.url
    try:
        response.raise_for_status()
    except ReadTimeout as e:
        msg = "The connection was successful, "
        msg += f"however the API call to {url} timed out after {timeout[1]}s "
        msg += "while waiting for the server to return data."
        raise APIError(msg) from e
    except HTTPError as e:
        msg = f"The API call to {url} failed with status code {response.status_code}."
        msg += "\nPerhaps your account credentials need to be refreshed?"
        raise APIError(msg) from e
    except (ConnectionError, Timeout) as e:
        msg = f"The API call to {url} failed due to connection issues."
        raise APIError(msg) from e
    except ProtocolError as e:
        msg = f"Did not receive any response for the API call to {url}."
        raise APIError(msg) from e
    except Exception as e:
        msg = "Unknown error."
        raise APIError(msg) from e

    return response


def handle_api_response(
    url: str,
    auth: tuple | None = None,
    params: dict[str, Any] | None = None,
    headers: dict[str, Any] | None = None,
    timeout: tuple = (3.05, 60 * 30),
    method: Literal["GET", "POST", "DELETE"] = "GET",
    data: str | None = None,
    verify: bool = True,
) -> requests.models.Response:
    """Handle an HTTP response.

    Apply retries and handle some common response codes.

    Args:
        url (str): The URL which trying to connect.
        auth (tuple, optional): A tuple of (username, password) for basic
            authentication. Defaults to None.
        params (Dict[str, Any], optional): The request parameters. Defaults to None.
        headers (Dict[str, Any], optional): The request headers. Defaults to None.
        method (Literal["GET", "POST", "DELETE"], optional): The HTTP method to use for
            the request. Defaults to "GET".
        timeout (tuple, optional): A tuple of (connect_timeout, read_timeout) in
            seconds. Defaults to (3.05, 60 * 30).
        data (str, optional): The request body data as a string. Defaults to None.
        verify (bool, optional): Whether to verify certificates. Defaults to True.

    Raises:
        - ConnectionError: If the connection timed out, as specified in `timeout`.
        - ReadTimeout: If the read timed out, as specified in `timeout`.
        - HTTPError: If the API raised an HTTPError.
        - APIError: Generic API error.

    Returns:
        requests.models.Response
    """
    response = handle_api_request(
        url=url,
        auth=auth,
        params=params,
        headers=headers,
        timeout=timeout,
        method=method,
        data=data,
        verify=verify,
    )

    return _handle_response(response)


def get_sql_server_table_dtypes(
    table: str, con: pyodbc.Connection, schema: str | None = None
) -> dict:
    """Get column names and types from a SQL Server database table.

    Args:
        table (str): The table for which to fetch dtypes.
        con (pyodbc.Connection): The connection to the database where the table
            is located.
        schema (str, optional): The schema where the table is located. Defaults
            to None.

    Returns:
        dict: A dictionary of the form {column_name: dtype, ...}.
    """
    query = f"""
    SELECT
        col.name,
        t.name,
        col.max_length
    FROM sys.tables AS tab
        INNER JOIN sys.columns AS col
            ON tab.object_id = col.object_id
        LEFT JOIN sys.types AS t
            ON col.user_type_id = t.user_type_id
    WHERE tab.name = '{table}'
    AND schema_name(tab.schema_id) = '{schema}'
    ORDER BY column_id;
    """
    cursor = con.cursor()
    query_result = cursor.execute(query).fetchall()
    cursor.close()

    dtypes = {}
    for row in query_result:
        column_name = row[0]
        dtype = row[1]
        length = row[2]
        if dtype == "varchar":
            dtypes[column_name] = dtype + f"({length})"
        else:
            dtypes[column_name] = dtype

    return dtypes


def cast_df_cols(
    df: pd.DataFrame,
    types_to_convert: list[Literal["datetime", "bool", "int", "object"]] | None = None,
) -> pd.DataFrame:
    """Cast the data types of columns in a DataFrame.

    Args:
        df (pd.DataFrame): The input DataFrame.
        types_to_convert (Literal[datetime, bool, int, object], optional): List of types
            to be converted. Defaults to ["datetime", "bool", "int"].

    Returns:
        pd.DataFrame: A DataFrame with modified data types.
    """
    if not types_to_convert:
        types_to_convert = ["datetime", "bool", "int"]

    df = df.replace({"False": False, "True": True})

    datetime_cols = (col for col, dtype in df.dtypes.items() if dtype.kind == "M")
    bool_cols = (col for col, dtype in df.dtypes.items() if dtype.kind == "b")
    int_cols = (col for col, dtype in df.dtypes.items() if dtype.kind == "i")
    object_cols = (col for col, dtype in df.dtypes.items() if dtype.kind == "O")

    if "datetime" in types_to_convert:
        for col in datetime_cols:
            df[col] = df[col].dt.strftime("%Y-%m-%d %H:%M:%S+00:00")
    if "bool" in types_to_convert:
        for col in bool_cols:
            df[col] = df[col].astype(pd.Int64Dtype())
    if "int" in types_to_convert:
        for col in int_cols:
            df[col] = df[col].astype(pd.Int64Dtype())
    if "object" in types_to_convert:
        for col in object_cols:
            df[col] = df[col].astype(pd.StringDtype())

    return df


def build_merge_query(
    table: str,
    primary_key: str,
    source: "Source",
    stg_schema: str | None = None,
    stg_table: str = "stg",
    schema: str | None = None,
    df: Optional["spark.DataFrame"] = None,
) -> str:
    """Build a merge query for the simplest possible upsert scenario.

    Used for:
    - updating and inserting all fields
    - merging on a single column, which has the same name in both tables

    Args:
        table (str): The table to merge into.
        primary_key (str): The column on which to merge.
        source (Source): The Databricks object used to connect to the Spark cluster on
            which the query will be executed.
        stg_schema (str, Optional): The schema where the staging table is located.
        stg_table (str, Optional): The table with new/updated data.
        schema (str, Optional): The schema where the table is located.
        df (spark.DataFrame, Optional): A Spark DataFrame whose data will be upserted
            to a table.
    """
    fqn = f"{schema}.{table}"
    stg_fqn = f"{stg_schema}.{stg_table}" if stg_schema else {stg_table}

    if schema is None:
        schema = source.DEFAULT_SCHEMA

    # The `DataFrame` *is* the staging table.
    if df:
        df.createOrReplaceTempView(stg_fqn)

    # Get column names
    columns_query_result = _get_table_columns(schema=schema, table=table, source=source)
    columns = list(columns_query_result)

    columns_stg_fqn = [f"{stg_table}.{col}" for col in columns]

    # Build merge query
    update_pairs = [f"existing.{col} = {stg_table}.{col}" for col in columns]
    return f"""
    MERGE INTO {fqn} existing
        USING {stg_fqn} {stg_table}
        ON {stg_table}.{primary_key} = existing.{primary_key}
        WHEN MATCHED
            THEN UPDATE SET {", ".join(update_pairs)}
        WHEN NOT MATCHED
            THEN INSERT({", ".join(columns)})
            VALUES({", ".join(columns_stg_fqn)});
    """


def _get_table_columns(schema: str, table: str, source: "Source") -> str:
    if source.__class__.__name__ == "Databricks":
        result = source.run(f"SHOW COLUMNS IN {schema}.{table}", "pandas")
        columns_query_result = result["col_name"].values
    else:
        columns_query = f"""
        SELECT
            col.name
        FROM sys.tables AS tab
            INNER JOIN sys.columns AS col
                ON tab.object_id = col.object_id
        WHERE tab.name = '{table}'
        AND schema_name(tab.schema_id) = '{schema}'
        ORDER BY column_id;
        """
        cursor = source.con.cursor()
        columns_query_result = cursor.execute(columns_query).fetchall()
        cursor.close()

    return columns_query_result


def gen_bulk_insert_query_from_df(
    df: pd.DataFrame, table_fqn: str, chunksize: int = 1000, **kwargs
) -> str:
    """Converts a DataFrame to a bulk INSERT query.

    Args:
        df (pd.DataFrame): The DataFrame which data should be put into the INSERT query.
        table_fqn (str): The fully qualified name (schema.table) of the table
            to be inserted into.

    Returns:
        str: A bulk insert query that will insert all data from `df` into `table_fqn`.

    Examples:
    >>> data = [(1, "_suffixnan", 1), (2, "Noneprefix", 0), (3, "fooNULLbar", 1, 2.34)]
    >>> df = pd.DataFrame(data, columns=["id", "name", "is_deleted", "balance"])
    >>> df
       id        name  is_deleted  balance
    0   1  _suffixnan           1      NaN
    1   2  Noneprefix           0      NaN
    2   3  fooNULLbar           1     2.34
    >>> query = gen_bulk_insert_query_from_df(
    >>>     df, table_fqn="users", status="APPROVED", address=None
    >>> )
    >>> print(query)
    INSERT INTO users (id, name, is_deleted, balance, status, address)
    VALUES (1, '_suffixnan', 1, NULL, 'APPROVED', NULL),
           (2, 'Noneprefix', 0, NULL, 'APPROVED', NULL),
           (3, 'fooNULLbar', 1, 2.34, 'APPROVED', NULL);
    """
    if df.shape[1] == 1:
        msg = "Currently, this function only handles DataFrames with at least two columns."
        raise NotImplementedError(msg)

    def _gen_insert_query_from_records(records: list[tuple]) -> str:
        tuples = map(str, tuple(records))

        # Change Nones to NULLs
        none_nan_pattern = r"(?<=\W)(nan|None)(?=\W)"
        values = re.sub(none_nan_pattern, "NULL", (",\n" + " " * 7).join(tuples))

        # Change the double quotes into single quotes, as explained above.
        # Note this pattern should be improved at a later time to cover more edge cases.
        double_quotes_pattern = r'(")(.*)(")(\)|,)'
        values_clean = re.sub(double_quotes_pattern, r"'\2'\4", values)
        # Hacky - replaces starting and ending double quotes.
        values_clean = (
            values.replace('",', "',").replace(', "', ", '").replace('("', "('")
        )

        return f"INSERT INTO {table_fqn} ({columns})\n\nVALUES {values_clean}"

    df = df.copy().assign(**kwargs)
    df = cast_df_cols(df)

    columns = ", ".join(df.columns)

    tuples_raw = df.itertuples(index=False, name=None)
    # Escape values with single quotes inside by adding another single quote
    # ("val'ue" -> "val''ue").
    # As Python wraps such strings in double quotes, which are interpreted as
    # column names by SQL Server, we later also replace the double quotes with
    # single quotes.
    tuples_escaped = [
        tuple(
            f"""{value.replace("'", "''")}""" if isinstance(value, str) else value
            for value in row
        )
        for row in tuples_raw
    ]

    if len(tuples_escaped) > chunksize:
        insert_query = ""
        chunk_start = 0
        for chunk_end in range(chunksize, len(tuples_escaped), chunksize):
            chunk = tuples_escaped[chunk_start:chunk_end]
            chunk_start += chunksize
            if len(tuples_escaped) - chunk_end < chunksize:
                chunk = tuples_escaped[chunk_end:]
            chunk_insert_query = _gen_insert_query_from_records(chunk)
            insert_query += chunk_insert_query + ";\n\n"
        return insert_query
    return _gen_insert_query_from_records(tuples_escaped)


def handle_if_empty(
    if_empty: Literal["warn", "skip", "fail"] = "warn",
    message: str | None = None,
    logger: logging.Logger | None = None,
) -> None:
    """Task for handling empty input.

    Args:
        if_empty (Literal, optional): Action to take when input is empty.
            Options are "warn", "skip", or "fail". Defaults to "warn".
        message (str, optional): Message to show in warnings or errors.
            Defaults to None.
        logger (logging.Logger, optional): Logger instance to use for warnings.
            If None, a default logger is created.

    Raises:
        ValueError: If `if_empty` is set to `fail`.
        SKIP: If `if_empty` is set to `skip`.
    """
    if not logger:
        logger = logging.getLogger(__name__)

    allowed = ["warn", "skip", "fail"]
    if if_empty not in allowed:
        error_msg = (
            f"Invalid value for if_empty: {if_empty}. Allowed values are {allowed}."
        )
        raise ValueError(error_msg)

    if if_empty == "warn":
        logger.warning(message)
    elif if_empty == "skip":
        raise SKIP(message)
    elif if_empty == "fail":
        raise ValueError(message)


def remove_newlines_and_tabs(df: pd.DataFrame) -> pd.DataFrame:
    """Remove some common data corruption from a pandas DataFrame.

    Args:
        df (pd.DataFrame): The pandas DataFrame to be cleaned up.

    Returns:
        pd.DataFrame: The cleaned up DataFrame
    """
    return df.replace(r"\n|\t", "", regex=True)


def call_shell(command: str) -> str:
    """Run a shell command and return the output."""
    try:
        result = subprocess.check_output(command, shell=True)  # noqa: S602
    except subprocess.CalledProcessError as e:
        # TODO: read the error message from stdout and pass here
        msg = "Generating the file failed."
        raise ValueError(msg) from e
    else:
        return result


def df_snakecase_column_names(df: pd.DataFrame) -> pd.DataFrame:
    """Snakecase the column names of a DataFrame.

    Args:
        df (pd.DataFrame): The DataFrame to snakecase.

    Returns:
        pd.DataFrame: The DataFrame with snakecased column names.
    """
    df.columns = (
        df.columns.str.strip().str.replace(" ", "_").str.replace("-", "_").str.lower()
    )
    return df


def join_dfs(
    df_left: pd.DataFrame,
    df_right: pd.DataFrame,
    left_on: str,
    right_on: str,
    columns_from_right_df: list[str] | None = None,
    how: Literal["left", "right", "outer", "inner", "cross"] = "left",
) -> pd.DataFrame:
    """Combine Data Frames according to the chosen method.

    Args:
        df_left (pd.DataFrame): Left dataframe.
        df_right (pd.DataFrame): Right dataframe.
        left_on (str): Column or index level names to join on in the left DataFrame.
        right_on (str): Column or index level names to join on in the right DataFrame.
        columns_from_right_df (list[str], optional): List of column to get from right
            dataframe. Defaults to None.
        how (Literal["left", "right", "outer", "inner", "cross"], optional): Type of
            merge to be performed. Defaults to "left".

    Returns:
        pd.DataFrame: Final dataframe after merging.
    """
    if columns_from_right_df is None:
        columns_from_right_df = df_right.columns

    return df_left.merge(
        df_right[columns_from_right_df],
        left_on=left_on,
        right_on=right_on,
        how=how,
    )


def add_viadot_metadata_columns(func: Callable) -> Callable:
    """A decorator for the 'to_df()' method.

    Adds viadot metadata columns to the returned DataFrame.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> pd.DataFrame:
        df = func(*args, **kwargs)

        # Accessing instance
        instance = args[0]
        _viadot_source = instance.__class__.__name__
        df["_viadot_source"] = _viadot_source
        df["_viadot_downloaded_at_utc"] = datetime.now(timezone.utc).replace(
            microsecond=0
        )
        return df

    return wrapper


def get_fqn(table_name: str, schema_name: str | None = None) -> str:
    """Get the fully qualified name of a table."""
    return f"{schema_name}.{table_name}" if schema_name else table_name


def validate_column_size(
    df: pd.DataFrame,
    tests: dict[str, Any],
    logger: logging.Logger,
    stream_level: int,
    failed_tests: int,
    failed_tests_list: list,
) -> tuple[int, list[str]]:
    """Validate the size of the columns in the DataFrame.

    Logic: TODO

    Args:
        df (pd.DataFrame): The pandas DataFrame to validate.
        tests (dict[str, Any]): A dictionary containing validation parameters.
            Expected format: `column_size`: dict{column: size}
        logger (logging.Logger): The logger to use.
        stream_level (int): The logging level to use for logging.
        failed_tests (int): Counter for the number of failed tests.
        failed_tests_list (list): List to store descriptions of failed tests.

    Returns:
        tuple[int, list[str]]: Updated count of failed tests and list of failure
        descriptions.

    """
    try:
        for k, v in tests["column_size"].items():
            column_max_length = (
                df.astype(str).apply(lambda s: s.str.len()).max().to_dict()
            )
            try:
                if v == column_max_length[k]:
                    logger.log(level=stream_level, msg="[column_size] for {k} passed.")
                else:
                    logger.log(
                        level=stream_level,
                        msg=f"[column_size] test for {k} failed. field length is different than {v}",
                    )
                    failed_tests += 1
                    failed_tests_list.append("column_size error")
            except Exception as e:
                logger.log(level=stream_level, msg=f"{e}")
    except TypeError:
        logger.log(
            level=stream_level,
            msg=f"Please provide `column_size` parameter as dictionary {'columns': value}.",
        )
    return failed_tests, failed_tests_list


def validate_column_unique_values(
    df: pd.DataFrame,
    tests: dict[str, Any],
    logger: logging.Logger,
    stream_level: int,
    failed_tests: int,
    failed_tests_list: list,
) -> tuple[int, list[str]]:
    """Validate whether a DataFrame column only contains unique values.

    Args:
        df (pd.DataFrame): The pandas DataFrame to validate.
        tests (dict[str, Any]): A dictionary containing validation parameters.
            Expected format: `column_unique_values`: list[columns]
        logger (logging.Logger): The logger to use.
        stream_level (int): The logging level to use for logging.
        failed_tests (int): Counter for the number of failed tests.
        failed_tests_list (list): List to store descriptions of failed tests.

    Returns:
        tuple[int, list[str]]: Updated count of failed tests and list of failure
        descriptions.

    """
    for column in tests["column_unique_values"]:
        df_size = df.shape[0]
        if df[column].nunique() == df_size:
            logger.log(
                level=stream_level,
                msg=f"[column_unique_values] Values are unique for {column} column.",
            )
        else:
            failed_tests += 1
            failed_tests_list.append("column_unique_values error")
            logger.log(
                level=stream_level,
                msg=f"[column_unique_values] Values for {column} are not unique.",
            )
    return failed_tests, failed_tests_list


def validate_column_list_to_match(
    df: pd.DataFrame,
    tests: dict[str, Any],
    logger: logging.Logger,
    stream_level: int,
    failed_tests: int,
    failed_tests_list: list,
) -> tuple[int, list[str]]:
    """Validate whether the columns of the DataFrame match the expected list.

    Args:
        df (pd.DataFrame): The pandas DataFrame to validate.
        tests (dict[str, Any]): A dictionary containing validation parameters.
            Expected format: `column_list_to_match`: list[columns]
        logger (logging.Logger): The logger to use.
        stream_level (int): The logging level to use for logging.
        failed_tests (int): Counter for the number of failed tests.
        failed_tests_list (list): List to store descriptions of failed tests.

    Returns:
        tuple[int, list[str]]: Updated count of failed tests and list of failure
        descriptions.

    """
    if set(tests["column_list_to_match"]) == set(df.columns):
        logger.log(level=stream_level, msg=f"{tests['column_list_to_match']} passed.")
    else:
        failed_tests += 1
        failed_tests_list.append("column_list_to_match error")
        logger.log(
            level=stream_level,
            msg="[column_list_to_match] failed. Columns are different than expected.",
        )
    return failed_tests, failed_tests_list


def validate_dataset_row_count(
    df: pd.DataFrame,
    tests: dict[str, Any],
    logger: logging.Logger,
    stream_level: int,
    failed_tests: int,
    failed_tests_list: list,
) -> tuple[int, list[str]]:
    """Validate the DataFrame row count.

    Args:
        df (pd.DataFrame): The pandas DataFrame to validate.
        tests (dict[str, Any]): A dictionary containing validation parameters.
            Expected format: `dataset_row_count`: dict: {'min': number, 'max', number}
        logger (logging.Logger): The logger to use.
        stream_level (int): The logging level to use for logging.
        failed_tests (int): Counter for the number of failed tests.
        failed_tests_list (list): List to store descriptions of failed tests.

    Returns:
        tuple[int, list[str]]: Updated count of failed tests and list of failure
        descriptions.

    """
    row_count = len(df.iloc[:, 0])
    max_value = tests["dataset_row_count"]["max"] or 100_000_000
    min_value = tests["dataset_row_count"]["min"] or 0

    if (row_count > min_value) and (row_count < max_value):
        logger.log(level=stream_level, msg="[dataset_row_count] passed.")
    else:
        failed_tests += 1
        failed_tests_list.append("dataset_row_count error")
        logger.log(
            level=stream_level,
            msg=f"[dataset_row_count] Row count ({row_count}) is not between {min_value} and {max_value}.",
        )
    return failed_tests, failed_tests_list


def validate_column_match_regex(
    df: pd.DataFrame,
    tests: dict[str, Any],
    logger: logging.Logger,
    stream_level: int,
    failed_tests: int,
    failed_tests_list: list,
) -> tuple[int, list[str]]:
    """Validate whether the values of a column match a regex pattern.

    Logic: TODO

    Args:
        df (pd.DataFrame): The pandas DataFrame to validate.
        tests (dict[str, Any]): A dictionary containing validation parameters.
            Expected format: `column_match_regex`: dict: {column: 'regex'}
        logger (logging.Logger): The logger to use.
        stream_level (int): The logging level to use for logging.
        failed_tests (int): Counter for the number of failed tests.
        failed_tests_list (list): List to store descriptions of failed tests.

    Returns:
        tuple[int, list[str]]: Updated count of failed tests and list of failure
        descriptions.

    """
    for k, v in tests["column_match_regex"].items():
        try:
            matches = df[k].apply(lambda x: bool(re.match(v, str(x))))  # noqa: B023
            if all(matches):
                logger.log(
                    level=stream_level,
                    msg=f"[column_match_regex] on {k} column passed.",
                )
            else:
                failed_tests += 1
                failed_tests_list.append("column_match_regex error")
                logger.log(
                    level=stream_level,
                    msg=f"[column_match_regex] on {k} column failed!",
                )
        except Exception as e:
            failed_tests += 1
            failed_tests_list.append("column_match_regex error")
            logger.log(
                level=stream_level,
                msg=f"[column_match_regex] Error in {k} column: {e}",
            )
    return failed_tests, failed_tests_list


def validate_column_sum(
    df: pd.DataFrame,
    tests: dict[str, Any],
    logger: logging.Logger,
    stream_level: int,
    failed_tests: int,
    failed_tests_list: list,
) -> tuple[int, list[str]]:
    """Validate the sum of a column in the DataFrame.

    Args:
        df (pd.DataFrame): The pandas DataFrame to validate.
        tests (dict[str, Any]): A dictionary containing validation parameters.
            Expected format: `column_sum`: dict: {column: {'min': number,
              'max': number}}
        logger (logging.Logger): The logger to use.
        stream_level (int): The logging level to use for logging.
        failed_tests (int): Counter for the number of failed tests.
        failed_tests_list (list): List to store descriptions of failed tests.

    Returns:
        tuple[int, list[str]]: Updated count of failed tests and list of failure
        descriptions.

    """
    for column, bounds in tests["column_sum"].items():
        col_sum = df[column].sum()
        min_bound = bounds["min"]
        max_bound = bounds["max"]
        if min_bound <= col_sum <= max_bound:
            logger.log(
                level=stream_level,
                msg=f"[column_sum] Sum of {col_sum} for {column} is within the expected range.",
            )
        else:
            failed_tests += 1
            failed_tests_list.append("column_sum error")
            logger.log(
                level=stream_level,
                msg=f"[column_sum] Sum of {col_sum} for {column} is out of the expected range - <{min_bound}:{max_bound}>",
            )
    return failed_tests, failed_tests_list


def validate(
    df: pd.DataFrame,
    tests: dict | None = None,
    stream_level: int = logging.INFO,
    logger: logging.Logger | None = None,
) -> None:
    """Validate data. All numbers in the ranges are inclusive.

    Available tests:
        - `column_size`: dict{column: size}
        - `column_unique_values`: list[columns]
        - `column_list_to_match`: list[columns]
        - `dataset_row_count`: dict: {'min': number, 'max', number}
        - `column_match_regex`: dict: {column: 'regex'}
        - `column_sum`: dict: {column: {'min': number, 'max': number}}

    Args:
        df (pd.DataFrame): The dataframe to validate.
        tests (dict, optional): Tests to apply on the data frame. Defaults to None.

    Raises:
        ValidationError: If validation failed for at least one test.
    """
    if logger is None:
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger(__name__)

    failed_tests = 0
    failed_tests_list = []

    if tests is not None:
        if "column_size" in tests:
            failed_tests, failed_tests_list = validate_column_size(
                df, tests, logger, stream_level, failed_tests, failed_tests_list
            )

        if "column_unique_values" in tests:
            failed_tests, failed_tests_list = validate_column_unique_values(
                df, tests, logger, stream_level, failed_tests, failed_tests_list
            )

        if "column_list_to_match" in tests:
            failed_tests, failed_tests_list = validate_column_list_to_match(
                df, tests, logger, stream_level, failed_tests, failed_tests_list
            )

        if "dataset_row_count" in tests:
            failed_tests, failed_tests_list = validate_dataset_row_count(
                df, tests, logger, stream_level, failed_tests, failed_tests_list
            )

        if "column_match_regex" in tests:
            failed_tests, failed_tests_list = validate_column_match_regex(
                df, tests, logger, stream_level, failed_tests, failed_tests_list
            )

        if "column_sum" in tests:
            failed_tests, failed_tests_list = validate_column_sum(
                df, tests, logger, stream_level, failed_tests, failed_tests_list
            )

    if failed_tests > 0:
        failed_tests_msg = ", ".join(failed_tests_list)
        msg = f"Validation failed for {failed_tests} test(s): {failed_tests_msg}"
        raise ValidationError(msg)


def validate_and_reorder_dfs_columns(
    dataframes_list: list[pd.DataFrame],
) -> list[pd.DataFrame]:
    """Validate if dataframes from the list have the same column structure.

    Reorder columns to match the first DataFrame if necessary.

    Args:
        dataframes_list (list[pd.DataFrame]): List containing DataFrames.

    Raises:
        IndexError: If the list of DataFrames is empty.
        ValueError: If DataFrames have different column structures.
    """
    if not dataframes_list:
        message = "The list of dataframes is empty."
        raise IndexError(message)

    first_df_columns = dataframes_list[0].columns

    # Check that all DataFrames have the same columns
    for i, df in enumerate(dataframes_list):
        if set(df.columns) != set(first_df_columns):
            message = f"""DataFrame at index {i} does not have the same structure as
            the first DataFrame."""
            raise ValueError(message)
        if not df.columns.equals(first_df_columns):
            # Reorder columns for DataFrame at index 'i' to match the first DataFrame.
            dataframes_list[i] = df.loc[:, first_df_columns]

    return dataframes_list


def skip_test_on_missing_extra(source_name: str, extra: str) -> None:
    """Skip all tests in a file when a required extra is not installed.

    Args:
        source_name (str): The name of the source for which dependencies are missing.
        extra (str): The name of the extra that is missing.
    """
    import pytest

    msg = f"Missing required extra '{extra}' for source '{source_name}'."
    pytest.skip(
        msg,
        allow_module_level=True,
    )


def filter_df_columns(data_frame: pd.DataFrame, columns: list) -> pd.DataFrame:
    """Function for creating DataFrame with only specified columns.

    Returns:
        pd.DataFrame: Pandas DataFrame
    """
    # Converting data_frame columns to lowercase for comparison
    data_frame_columns = [col.casefold() for col in data_frame.columns]
    needed_columns = []
    non_available_columns = []

    for column in columns:
        # Converting user-specified column to lowercase for comparison
        column_lower = column.casefold()

        if column_lower in data_frame_columns:
            # Find the original column name from data_frame.columns
            original_column_name = next(
                col for col in data_frame.columns if col.casefold() == column_lower
            )
            needed_columns.append(original_column_name)
        else:
            non_available_columns.append(column)

    # Error logger
    if non_available_columns:
        logging.error(
            f"Name of the columns: '{' | '.join(non_available_columns)}' are not in DataFrame. Please check spelling!\n"
            f"Available columns: {' | '.join(data_frame.columns)}"
        )
        msg = "Provided columns are not available!"
        raise ValueError(msg)

    # Selecting only needed columns from the original DataFrame
    return data_frame.loc[:, needed_columns]


def anonymize_df(
    df: pd.DataFrame,
    columns: list[str],
    method: Literal["mask", "hash"] = "mask",
    mask_value: str = "***",
    date_column: str | None = None,
    days: int | None = None,
    logger: logging.Logger | None = None,
) -> pd.DataFrame:
    """Function that anonymize data in the dataframe in selected columns.

    It is possible to specify the condtition, for which data older than specified
    number of days will be anonymized.

    Args:
        df (pd.DataFrame): Dataframe with data to anonymize.
        columns (list[str]): List of columns to anonymize.
        method (Literal["mask", "hash"], optional): Method of anonymizing data.
            "mask" -> replace the data with "value" arg.
            "hash" -> replace the data with the hash value of an object
            (using `hash()` method). Defaults to "mask".
        mask_value (str, optional): Value to replace the data with  when using the
            "mask" method. Defaults to "***".
        date_column (str, optional): Name of the date column used to identify
            rows that are older than a specified number of days. Defaults to None.
        days (int, optional): The number of days beyond which we want to
            anonymize the data, e.g. older that 2 years can be: 2*365. Defaults to None.

    Examples:
        1. Implement "mask" method with "***" for all data in columns:
            ["email", "last_name", "phone"]:
            >>> anonymize_df(df=df, columns=["email", "last_name", "phone"])
        2. Implement "hash" method with in columns: ["email", "last_name", "phone"]:
            >>> anonymize_df(df=df, columns=["email", "last_name", "phone"],
            method = "hash")
        3. Implement "mask" method with "***" for data in columns:
            ["email", "last_name", "phone"], that is older than two years in
            "submission_date" column:
            >>> anonymize_df(df=df, columns=["email", "last_name", "phone"],
            date_column="submission_date", days=2*365)

    Raises:
        ValueError: If method or columns not found.

    Returns:
        pd.DataFrame: Operational dataframe with anonymized data.
    """
    if logger is None:
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger(__name__)

    if all(col not in df.columns for col in columns):
        msg = f"""At least one of the following columns is not found in dataframe:
            {columns} or argument is not list. Provide list with proper column names."""
        raise ValueError(msg)

    if days and date_column:
        days_ago = datetime.now().date() - timedelta(days=days)
        df["temp_date_col"] = pd.to_datetime(df[date_column]).dt.date

        to_hash = df["temp_date_col"] < days_ago
        if any(to_hash) is False:
            logger.warning(f"No data that is older than {days} days.")
        else:
            logger.info(
                f"Data older than {days} days in {columns} columns will be anonymized."
            )
    else:
        to_hash = len(df.index) * [True]
        logger.info(
            f"""The 'days' and 'date_column' arguments were not specified. All data in
            {columns} columns will be anonymized."""
        )

    if method == "mask":
        df.loc[to_hash, columns] = mask_value
    elif method == "hash":
        df.loc[to_hash, columns] = df.loc[to_hash, columns].apply(
            lambda x: x.apply(hash)
        )
    else:
        msg = "Method not found. Use one of the available methods: 'mask', 'hash'."
        raise ValueError(msg)

    df.drop(columns=["temp_date_col"], inplace=True, errors="ignore")
    return df


def df_converts_bytes_to_int(df: pd.DataFrame) -> pd.DataFrame:
    """Task to convert bytes values to int.

    Args:
        df (pd.DataFrame): Data Frame to convert

    Returns:
        pd.DataFrame: Data Frame after convert
    """
    return df.applymap(lambda x: int(x) if isinstance(x, bytes) else x)


def df_clean_column(
    df: pd.DataFrame, columns_to_clean: list[str] | None = None
) -> pd.DataFrame:
    """Remove special characters from a pandas DataFrame.

    Args:
    df (pd.DataFrame): The DataFrame to clean.
    columns_to_clean (List[str]): A list of columns to clean. Defaults is None.

    Returns:
    pd.DataFrame: The cleaned DataFrame
    """
    df = df.copy()

    if columns_to_clean is None:
        df.replace(
            to_replace=[r"\\t|\\n|\\r", "\t|\n|\r"],
            value=["", ""],
            regex=True,
            inplace=True,
        )
    else:
        for col in columns_to_clean:
            df[col].replace(
                to_replace=[r"\\t|\\n|\\r", "\t|\n|\r"],
                value=["", ""],
                regex=True,
                inplace=True,
            )
    return df


def parse_dates(
    date_filter: str | tuple[str, str] | None = None,
    dynamic_date_symbols: list[str] | None = None,
    dynamic_date_format: str = "%Y-%m-%d",
    dynamic_date_timezone: str = "UTC",
) -> pendulum.Date | tuple[pendulum.Date, pendulum.Date] | None:
    """Parses a date or date range, supporting dynamic date symbols.

    Args:
        date_filter (str | tuple[str, str] | None):
            - A single date string (e.g., "2024-03-03").
            - A tuple containing exactly two date strings, 'start' and 'end' date.
            - None, which applies no date filter.
            Defaults to None.
        dynamic_date_symbols (list[str]): Symbols for dynamic date handling.
            Defaults to None.
        dynamic_date_format (str): Format used for dynamic date parsing.
            Defaults to "%Y-%m-%d".
        dynamic_date_timezone (str): Timezone used for dynamic date processing.
            Defaults to "UTC".

    Returns:
        pendulum.Date: If a single date is provided.
        tuple[pendulum.Date, pendulum.Date]: If a date range is provided.
        None: If `date_filter` is None.

    Raises:
        ValueError: If `date_filter` is neither a string nor a tuple of exactly
            two strings.
    """
    if date_filter is None:
        return None

    dynamic_date_symbols = dynamic_date_symbols or ["<<", ">>"]

    ddh = DynamicDateHandler(
        dynamic_date_symbols=dynamic_date_symbols,
        dynamic_date_format=dynamic_date_format,
        dynamic_date_timezone=dynamic_date_timezone,
    )

    match date_filter:
        case str():
            return pendulum.parse(ddh.process_dates(date_filter)).date()

        case (start, end) if isinstance(start, str) and isinstance(end, str):
            return (
                pendulum.parse(ddh.process_dates(start)).date(),
                pendulum.parse(ddh.process_dates(end)).date(),
            )

        case _:
            msg = "date_filter must be a string, a tuple of exactly 2 dates, or None."
            raise ValueError(msg)



"""Flows for downloading data from EntraID and uploading it to AWS Redshift Spectrum."""

from typing import Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import (
    df_to_redshift_spectrum,
    entraid_to_df,
)
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="extract--entraid--redshift_spectrum",
    description="Extract data from EntraID and load it into AWS Redshift Spectrum.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def entraid_to_redshift_spectrum(  # noqa: PLR0913
    to_path: str,
    schema_name: str,
    table: str,
    extension: str = ".parquet",
    if_exists: Literal["overwrite", "append"] = "overwrite",
    partition_cols: list[str] | None = None,
    index: bool = False,
    compression: Literal["snappy", "gzip", "zstd"] | None = None,
    sep: str = ",",
    aws_config_key: str | None = None,
    credentials_secret: str | None = None,
    entraid_credentials_secret: str | None = None,
    entraid_config_key: str | None = None,
) -> None:
    """Extract data from EntraID and load it into AWS Redshift Spectrum.

    This function downloads data from EntraID and uploads it to AWS Redshift Spectrum.

    Modes:
    It downloads data from EntraID and creates a table from it.


    Args:
        to_path (str): Path to a S3 folder where the table will be located. Defaults to
            None.
        schema_name (str): AWS Glue catalog database name.
        table (str): AWS Glue catalog table name.
        extension (str): Required file type. Accepted file formats are 'csv' and
            'parquet'.
        if_exists (Literal["overwrite", "append"], optional): Set to 'overwrite' to
          recreate the existing table.
        partition_cols (list[str], optional): List of column names that will be used to
            create partitions. Only takes effect if dataset=True. Defaults to None.
        index (bool, optional): Write row names (index). Defaults to False.
        compression (Literal["snappy", "gzip", "zstd"], optional): Compression method
        sep (str, optional): Field delimiter for the output file. Defaults to ','.
        aws_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        credentials_secret (str, optional): The name of a secret block in Prefect
            that stores AWS credentials. Defaults to None.
        entraid_credentials_secret (str, optional): The name of the secret storing
            EntraID credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        entraid_config_key (str, optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
    """
    df = entraid_to_df(
        credentials_secret=entraid_credentials_secret,
        config_key=entraid_config_key,
    )

    df_to_redshift_spectrum(
        df=df,
        to_path=to_path,
        schema_name=schema_name,
        table=table,
        extension=extension,
        if_exists=if_exists,
        partition_cols=partition_cols,
        index=index,
        compression=compression,
        sep=sep,
        config_key=aws_config_key,
        credentials_secret=credentials_secret,
    )



"""Flows for loading data from Sharepoint to Redshift Spectrum."""

from typing import Any, Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import (
    df_to_redshift_spectrum,
    sharepoint_list_to_df,
)
from viadot.orchestration.prefect.utils import with_flow_timeout_param


@flow(
    name="extract--sharepoint--list--redshift_spectrum",
    description="Extract data from Sharepoint Lists and load it into AWS Redshift Spectrum.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def sharepoint_list_to_redshift_spectrum(  # noqa: PLR0913
    to_path: str,
    schema_name: str,
    table: str,
    list_site: str,
    list_name: str,
    sep: str = ",",
    extension: str = ".parquet",
    if_exists: Literal["overwrite", "append"] = "overwrite",
    default_protocol: str | None = "https://",
    partition_cols: list[str] | None = None,
    index: bool = False,
    compression: str | None = None,
    aws_config_key: str | None = None,
    credentials_secret: str | None = None,
    credentials_secret_cert_auth: str | None = None,
    credentials_secret_cert_password: str | None = None,
    sharepoint_credentials_secret: str | None = None,
    sharepoint_config_key: str | None = None,
    query: str | None = None,
    select: list[str] | None = None,
    tests: dict[str, Any] | None = None,
) -> None:
    """Flow to load data from a SharePoint list to Redshift Spectrum.

    Args:
        to_path (str): The path where the data will be stored.
        schema_name (str): The name of the schema in Redshift Spectrum.
        table (str): The name of the table in Redshift Spectrum.
        list_site (str):  The Sharepoint site on which the list is stored.
        list_name (str): The name of the SharePoint list.
        sep (str, optional): The separator used in the file. Defaults to ",".
        extension (str, optional): The file extension to use. Defaults to ".parquet".
        default_protocol (str, optional): The default protocol to use for
                SharePoint URLs.Defaults to "https://".
        if_exists (Literal["overwrite", "append"], optional): Action if the table
            exists.Defaults to "overwrite".
        partition_cols (list[str] | None, optional): Columns to partition the data by.
            Defaults to None.
        index (bool, optional): Whether to include the index in the output.
            Defaults to False.
        compression (str | None, optional): Compression type to use. Defaults to None.
        aws_config_key (str | None, optional): AWS configuration key. Defaults to None.
        credentials_secret (str | None, optional): Name of the secret storing
            AWS credentials. Defaults to None.
        credentials_secret_cert_auth (str | None, optional): Name of the secret storing
            SharePoint credentials for certificate authentication. Defaults to None.
        credentials_secret_cert_password (str | None, optional): Name of the secret
            storing the password for the certificate file. Defaults to None.
        sharepoint_credentials_secret (str | None, optional): Name of the secret
            storing SharePoint credentials. Defaults to None.
        sharepoint_config_key (str | None, optional): Key in the config for
            SharePoint  credentials. Defaults to None.
        query (str | None, optional): Query to filter items in the SharePoint list.
            Defaults to None.
        select (list[str] | None, optional): Fields to include from the SharePoint list.
            Defaults to None.
        tests (dict[str, Any] | None, optional): Tests to validate the DataFrame.
            Defaults to None.

    Returns:
        None
    """
    # Tasks
    df = sharepoint_list_to_df(
        list_site=list_site,
        list_name=list_name,
        default_protocol=default_protocol,
        query=query,
        select=select,
        credentials_secret_basic_auth=sharepoint_credentials_secret,
        credentials_secret_cert_auth=credentials_secret_cert_auth,
        credentials_secret_cert_password=credentials_secret_cert_password,
        config_key=sharepoint_config_key,
        tests=tests,
    )
    df_to_redshift_spectrum(
        df=df,
        to_path=to_path,
        schema_name=schema_name,
        table=table,
        extension=extension,
        if_exists=if_exists,
        partition_cols=partition_cols,
        index=index,
        compression=compression,
        sep=sep,
        config_key=aws_config_key,
        credentials_secret=credentials_secret,
    )



"""Flows for downloading data from Informix and uploading it to Redshift Spectrum."""

from typing import Any, Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_redshift_spectrum, informix_to_df
from viadot.orchestration.prefect.utils import (
    DynamicDateHandler,
    with_flow_timeout_param,
)


@flow(
    name="extract--informix--redshift_spectrum",
    description="Extract data from Informix and load it into AWS Redshift Spectrum.",
    retries=1,
    retry_delay_seconds=60,
)
@with_flow_timeout_param()
def informix_to_redshift_spectrum(  # noqa: PLR0913
    query: str,
    to_path: str,
    schema_name: str,
    table: str,
    sep: str = ",",
    extension: str = ".parquet",
    if_exists: Literal["overwrite", "append"] = "overwrite",
    partition_cols: list[str] | None = None,
    index: bool = False,
    compression: str | None = None,
    dynamic_date_symbols: list[str] = ["<<", ">>"],  # noqa: B006
    dynamic_date_format: str = "%Y%m%d",
    dynamic_date_timezone: str = "UTC",
    tests: dict[str, Any] | None = None,
    config_key: str | None = None,
    credentials_secret: str | None = None,
    informix_credentials_secret: str | None = None,
    informix_config_key: str | None = None,
) -> None:
    """Flow to load data from a Informix source to Redshift Spectrum.

    Args:
        query (str): The query to execute.
        to_path (str): The path to the destination file.
        schema_name (str): The schema name.
        table (str): The table name.
        sep (str): The separator for the CSV file.
        extension (str): The extension for the file.
        if_exists (Literal["overwrite", "append"]): The action to take if
            the table already exists.
        partition_cols (list[str]): The columns to partition the data by.
        index (bool): Whether to include the index in the output file.
        compression (str): The compression type to use for the output file.
        dynamic_date_symbols (list[str]): The symbols used for dynamic date handling.
        dynamic_date_format (str): The format used for dynamic date parsing.
        dynamic_date_timezone (str): The timezone used for dynamic date processing.
        tests (dict[str, Any]): The tests to run on the data.
        config_key (str): The configuration key.
        credentials_secret (str): The credentials secret.
        informix_credentials_secret (str): The Informix credentials secret.
        informix_config_key (str): The Informix configuration key.
    """
    ddh = DynamicDateHandler(
        dynamic_date_symbols, dynamic_date_format, dynamic_date_timezone
    )
    query = ddh.process_dates(query)  # type: ignore

    df = informix_to_df(
        query=query,
        config_key=informix_config_key,
        credentials_secret=informix_credentials_secret,
        tests=tests,
    )

    df_to_redshift_spectrum(
        df=df,
        to_path=to_path,
        schema_name=schema_name,
        table=table,
        extension=extension,
        if_exists=if_exists,
        partition_cols=partition_cols,
        index=index,
        compression=compression,
        sep=sep,
        config_key=config_key,
        credentials_secret=credentials_secret,
    )



"""Task to download data from Mindful API into a Pandas DataFrame."""

from datetime import date
from typing import Literal

import pandas as pd
from prefect import get_run_logger, task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import Mindful


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=60 * 60)
def mindful_to_df(
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    region: Literal["us1", "us2", "us3", "ca1", "eu1", "au1"] = "eu1",
    endpoint: str | None = None,
    date_interval: list[date] | None = None,
    limit: int = 1000,
) -> pd.DataFrame:
    """Task to download data from Mindful API.

    Args:
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret where credentials are stored. Defaults to None.
        region (Literal[us1, us2, us3, ca1, eu1, au1], optional): Survey Dynamix region
            from where to interact with the mindful API. Defaults to "eu1" English
            (United Kingdom).
        endpoint (Optional[str], optional): API endpoint for an individual request.
            Defaults to None.
        date_interval (Optional[List[date]], optional): Date time range detailing the
            starting date and the ending date. If no range is passed, one day of data
            since this moment will be retrieved. Defaults to None.
        limit (int, optional): The number of matching interactions to return.
            Defaults to 1000.

    Examples:
        data_frame = mindful_to_df(
            config_key=config_key,
            azure_key_vault_secret=azure_key_vault_secret,
            region=region,
            endpoint=end,
            date_interval=date_interval,
            limit=limit,
        )

    Returns:
        pd.DataFrame: The response data as a pandas DataFrame.
    """
    logger = get_run_logger()

    if not (azure_key_vault_secret or config_key):
        raise MissingSourceCredentialsError

    if not config_key:
        credentials = get_credentials(azure_key_vault_secret)

    if endpoint is None:
        logger.warning(
            "The API endpoint parameter was not defined. The default value is 'surveys'."
        )
        endpoint = "surveys"

    mindful = Mindful(
        credentials=credentials,
        config_key=config_key,
        region=region,
    )
    mindful.api_connection(
        endpoint=endpoint,
        date_interval=date_interval,
        limit=limit,
    )

    return mindful.to_df()



"""SFTP connector."""

from io import BytesIO, StringIO
from pathlib import Path
import re
from stat import S_ISDIR, S_ISREG
import time
from typing import Literal

import pandas as pd
import paramiko
from paramiko.sftp import SFTPError
from pydantic import BaseModel

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns


class SftpCredentials(BaseModel):
    """Checking for values in SFTP credentials dictionary.

    Two key values are held in the Salesforce connector:
        - hostname: IP address of the SFTP server..
        - username: The user name for SFTP connection.
        - password: The passwrod for SFTP connection.
        - port: The port to use for the connection.
        - rsa_key: The SSH key to use for the connection. Only RSA is currently
            supported.

    Args:
        BaseModel (pydantic.main.ModelMetaclass): A base class for creating
            Pydantic models.
    """

    hostname: str
    username: str
    password: str
    port: int
    rsa_key: str | None


class Sftp(Source):
    """Class implementing a SFTP server connection."""

    def __init__(
        self,
        *args,
        credentials: SftpCredentials | None = None,
        config_key: str = "sftp",
        **kwargs,
    ):
        """Create an instance of SFTP.

        Args:
            credentials (SftpCredentials, optional): SFTP credentials. Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to "sftp".

        Notes:
            self.conn is paramiko.SFTPClient.from_transport method that contains
            additional methods like get, put, open etc. Some of them were not
            implemented in that class. For more check documentation
            (https://docs.paramiko.org/en/stable/api/sftp.html).

            sftp = Sftp()
            sftp.conn.open(filename='folder_a/my_file.zip', mode='r')

        Raises:
            CredentialError: If credentials are not provided in viadot config or
                directly as a parameter.
        """
        credentials = credentials or get_source_credentials(config_key)

        if credentials is None:
            message = "Missing credentials."
            raise CredentialError(message)

        validated_creds = dict(SftpCredentials(**credentials))
        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.conn = None
        self.hostname = validated_creds.get("hostname")
        self.username = validated_creds.get("username")
        self.password = validated_creds.get("password")
        self.port = validated_creds.get("port")
        self.rsa_key = validated_creds.get("rsa_key")

    def _get_file_object(self, file_name: str) -> BytesIO:
        """Copy a remote file from the SFTP server and write to a file-like object.

        Args:
            file_name (str, optional): File name to copy.

        Returns:
            BytesIO: file-like object.
        """
        file_object = BytesIO()
        try:
            self.conn.getfo(file_name, file_object)

        except FileNotFoundError as error:
            raise SFTPError from error

        else:
            return file_object

    def get_connection(self) -> paramiko.SFTPClient:
        """Returns a SFTP connection object.

        Returns: paramiko.SFTPClient.
        """
        ssh = paramiko.SSHClient()

        if not self.rsa_key:
            transport = paramiko.Transport((self.hostname, self.port))
            transport.connect(None, self.username, self.password)

            self.conn = paramiko.SFTPClient.from_transport(transport)

        else:
            mykey = paramiko.RSAKey.from_private_key(StringIO(self.rsa_key))
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())  # noqa: S507
            ssh.connect(self.hostname, username=self.username, pkey=mykey)
            time.sleep(1)
            self.conn = ssh.open_sftp()

        self.logger.info("Connected to the SFTP server.")

    @add_viadot_metadata_columns
    def to_df(
        self,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        file_name: str | None = None,
        sep: str = "\t",
        columns: list[str] | None = None,
    ) -> pd.DataFrame:
        r"""Copy a remote file from the SFTP server and write it to Pandas dataframe.

        Args:
            if_empty (Literal["warn", "skip", "fail"], optional): What to do if
                the fetch produces no data. Defaults to "warn".
            file_name (str, optional): The name of the file to download.
            sep (str, optional): The delimiter for the source file. Defaults to "\t".
            columns (list[str], optional): List of columns to select from file.
                Defaults to None.

        Returns:
            pd.DataFrame: The response data as a Pandas Data Frame plus viadot metadata.
        """
        byte_file = self._get_file_object(file_name=file_name)
        byte_file.seek(0)

        self._close_conn()

        suffix = Path(file_name).suffix
        if suffix == ".csv":
            df = pd.read_csv(byte_file, sep=sep, usecols=columns)

        elif suffix == ".parquet":
            df = pd.read_parquet(byte_file, usecols=columns)

        elif suffix == ".tsv":
            df = pd.read_csv(byte_file, sep=sep, usecols=columns)

        elif suffix in [".xls", ".xlsx", ".xlsm"]:
            df = pd.read_excel(byte_file, usecols=columns)

        elif suffix == ".json":
            df = pd.read_json(byte_file)

        elif suffix == ".pkl":
            df = pd.read_pickle(byte_file)  # noqa: S301

        elif suffix == ".sql":
            df = pd.read_sql(byte_file)

        elif suffix == ".hdf":
            df = pd.read_hdf(byte_file)

        else:
            message = (
                f"Unable to read file '{Path(file_name).name}', "
                + f"unsupported filetype: {suffix}"
            )
            raise ValueError(message)

        if df.empty:
            self._handle_if_empty(
                if_empty=if_empty,
                message="The response does not contain any data.",
            )
        else:
            self.logger.info("Successfully downloaded data from the SFTP server.")

        return df

    def _ls(self, path: str | None = ".", recursive: bool = False) -> list[str]:
        """List files in specified directory, with optional recursion.

        Args:
            path (str | None): Full path to the remote directory to list.
                Defaults to ".".
            recursive (bool): Whether to list files recursively. Defaults to False.

        Returns:
            list[str]: List of files in the specified directory.
        """
        files_list = []

        path = "." if path is None else path
        try:
            if not recursive:
                return [
                    str(Path(path) / attr.filename)
                    for attr in self.conn.listdir_attr(path)
                    if S_ISREG(attr.st_mode)
                ]

            for attr in self.conn.listdir_attr(path):
                full_path = str(Path(path) / attr.filename)
                if S_ISDIR(attr.st_mode):
                    files_list.extend(self._ls(full_path, recursive=True))
                else:
                    files_list.append(full_path)
        except FileNotFoundError as e:
            self.logger.info(f"Directory not found: {path}. Error: {e}")
        except Exception as e:
            self.logger.info(f"Error accessing {path}: {e}")

        return files_list

    def get_files_list(
        self,
        path: str | None = None,
        recursive: bool = False,
        matching_path: str | None = None,
    ) -> list[str]:
        """List files in `path`.

        Args:
            path (str | None): Destination path from where to get the structure.
                Defaults to None.
            recursive (bool): Get the structure in deeper folders.
                Defaults to False.
            matching_path (str | None): Filtering folders to return by a regex
                pattern. Defaults to None.

        Returns:
            list[str]: List of files in the specified path.
        """
        files_list = self._ls(path=path, recursive=recursive)

        if matching_path is not None:
            files_list = [f for f in files_list if re.match(matching_path, f)]

        self._close_conn()

        self.logger.info("Successfully loaded file list from SFTP server.")

        return files_list

    def _close_conn(self) -> None:
        """Close the SFTP server connection."""
        if self.conn is not None:
            self.conn.close()
            self.conn = None



from contextlib import nullcontext as does_not_raise

import pytest

from viadot.utils import skip_test_on_missing_extra


try:
    from viadot.sources import MinIO
except ImportError:
    skip_test_on_missing_extra(source_name="MinIO", extra="aws")


TEST_BUCKET = "spark"
TEST_SCHEMA = "test_schema"
TEST_SCHEMA_URL = f"s3a://{TEST_BUCKET}/{TEST_SCHEMA}"
TEST_SCHEMA_PATH = TEST_SCHEMA
TEST_TABLE = "test_table"
TEST_TABLE_URL = f"{TEST_SCHEMA_URL}/{TEST_TABLE}"
TEST_TABLE_PATH = f"{TEST_SCHEMA_PATH}/{TEST_TABLE}"
TEST_TABLE_FILE_PATH = f"{TEST_TABLE_PATH}/{TEST_TABLE}.parquet"


@pytest.fixture(scope="session")
def minio(minio_config_key):
    minio = MinIO(config_key=minio_config_key)
    minio.rm(TEST_TABLE_FILE_PATH)

    return minio


def test_check_connection(minio):
    with does_not_raise():
        minio.check_connection()


def test_from_df(minio, DF):
    # Assumptions.
    file_exists = minio._check_if_file_exists(TEST_TABLE_FILE_PATH)
    assert not file_exists

    # Test.
    minio.from_df(DF, schema_name=TEST_SCHEMA, table_name=TEST_TABLE)

    file_exists = minio._check_if_file_exists(TEST_TABLE_FILE_PATH)
    assert file_exists

    # Cleanup.
    minio.rm(TEST_TABLE_FILE_PATH)


def test_from_df_path(minio, DF):
    """Test that the `path` parameter in `from_df()` works."""
    # Assumptions.
    file_exists = minio._check_if_file_exists(TEST_TABLE_FILE_PATH)
    assert not file_exists

    # Test.
    minio.from_df(DF, path=TEST_TABLE_FILE_PATH)

    file_exists = minio._check_if_file_exists(TEST_TABLE_FILE_PATH)
    assert file_exists

    # Cleanup.
    minio.rm(TEST_TABLE_FILE_PATH)


def test_ls(minio, DF):
    # Assumptions.
    files = list(minio.ls(TEST_TABLE_PATH + "/"))
    assert TEST_TABLE_FILE_PATH not in files

    # Test.
    minio.from_df(DF, path=TEST_TABLE_FILE_PATH)
    files = list(minio.ls(TEST_TABLE_PATH + "/"))
    assert TEST_TABLE_FILE_PATH in files

    # Cleanup.
    minio.rm(TEST_TABLE_FILE_PATH)



from unittest.mock import Mock, patch

import pandas as pd
import pytest

from viadot.sources.business_core import BusinessCore


@pytest.fixture(scope="module")
def business_core():
    return BusinessCore(
        url="https://api.businesscore.ae/api/GetCustomerData",
        filters={
            "BucketCount": 10,
            "BucketNo": 1,
            "FromDate": None,
            "ToDate": None,
        },
        credentials={
            "username": "test",
            "password": "test123",  # pragma: allowlist secret
        },
    )


@patch("viadot.sources.business_core.handle_api_response")
def test_generate_token(mock_api_response, business_core):
    mock_api_response.return_value = Mock(text='{"access_token": "12345"}')
    token = business_core.generate_token()
    t = "12345"
    assert token == t


def test__clean_filters(business_core):
    filters_raw = {
        "BucketCount": 10,
        "BucketNo": 1,
        "FromDate": None,
        "ToDate": None,
    }
    filters_clean = business_core._clean_filters(filters_raw)
    assert filters_clean == {
        "BucketCount": 10,
        "BucketNo": 1,
        "FromDate": "&",
        "ToDate": "&",
    }


def test_to_df(business_core):
    with patch.object(
        business_core,
        "get_data",
        return_value=[{"id": 1, "name": "John Doe"}],
    ):
        df = business_core.to_df()
        assert isinstance(df, pd.DataFrame)
        assert len(df.columns) == 4
        assert len(df) == 1
        assert df["id"].tolist() == [1]
        assert df["name"].tolist() == ["John Doe"]



"""Tasks for interacting with Azure Data Lake (gen2)."""

import contextlib

import pandas as pd
from prefect import task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials


with contextlib.suppress(ImportError):
    from viadot.sources import AzureDataLake


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60)
def adls_upload(
    to_path: str,
    from_path: str | None = None,
    recursive: bool = False,
    overwrite: bool = False,
    credentials_secret: str | None = None,
    config_key: str | None = None,
) -> None:
    """Upload file(s) to Azure Data Lake.

    Credentials can be specified as either a key inside viadot config file,
    or the name of the Prefect `AzureKeyVaultSecretReference` block document
    storing the reference to an Azure Key Vault secret.

    Args:
        to_path (str, optional): The destination path.
        recursive (bool, optional): Set this to true if uploading entire directories.
            Defaults to False.
        from_path (str, optional): The local path from which to upload the file(s).
            Defaults to None.
        overwrite (bool, optional): Whether to overwrite files in the lake. Defaults
            to False.
        credentials_secret (str, optional): The name of the Azure Key Vault secret
            storing the credentials.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials.
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    credentials = get_credentials(credentials_secret)
    lake = AzureDataLake(credentials=credentials, config_key=config_key)

    lake.upload(
        from_path=from_path,
        to_path=to_path,
        recursive=recursive,
        overwrite=overwrite,
    )


@task(retries=3, retry_delay_seconds=10)
def df_to_adls(
    df: pd.DataFrame,
    path: str,
    sep: str = "\t",
    credentials_secret: str | None = None,
    config_key: str | None = None,
    overwrite: bool = False,
) -> None:
    r"""Upload a pandas `DataFrame` to a file on Azure Data Lake.

    Args:
        df (pd.DataFrame): The pandas DataFrame to upload.
        path (str): The destination path. Defaults to None.
        sep (str, optional): The separator to use in the `to_csv` function. Defaults to
            "\t".
        overwrite (bool, optional): Whether to overwrite files in the lake. Defaults
            to False.
        credentials_secret (str, optional): The name of the Azure Key Vault secret
            storing the credentials.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    credentials = get_credentials(credentials_secret)
    lake = AzureDataLake(credentials=credentials, config_key=config_key)

    lake.from_df(
        df=df,
        path=path,
        sep=sep,
        overwrite=overwrite,
    )



"""'test_mediatool.py'."""

import json
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from viadot.exceptions import CredentialError
from viadot.sources import Mediatool
from viadot.sources.mediatool import MediatoolCredentials


variables = {
    "credentials": {"user_id": "test_user", "token": "test_token"},
    "organizations": {
        "organizations": [{"_id": "1", "name": "Org1", "abbreviation": "O1"}]
    },
    "media_entries": {
        "mediaEntries": [
            {
                "_id": "1",
                "campaignId": "1",
                "organizationId": "1",
                "vehicleId": "1",
                "mediaTypeId": "1",
                "name": "Entry1",
                "customCol1": "CustomValue1",
                "customCol2": "CustomValue2",
            }
        ]
    },
    "vehicle": {"vehicle": {"_id": "1", "name": "Vehicle1"}},
    "campaigns": {
        "campaigns": [
            {"_id": "1", "name": "Campaign1", "conventionalName": "conventionalName1"}
        ]
    },
    "media_types": {"mediaType": {"_id": "1", "name": "Type1", "type": "Type"}},
}


@pytest.fixture
def mock_multiple_api_responses():
    """Fixture to mock multiple API responses for Mediatool.to_df() method."""
    mock_organizations_response = MagicMock()
    mock_organizations_response.text = json.dumps(variables["organizations"])

    mock_media_entries_response = MagicMock()
    mock_media_entries_response.text = json.dumps(variables["media_entries"])

    mock_vehicle_response = MagicMock()
    mock_vehicle_response.text = json.dumps(variables["vehicle"])

    mock_campaigns_response = MagicMock()
    mock_campaigns_response.text = json.dumps(variables["campaigns"])

    mock_media_types_response = MagicMock()
    mock_media_types_response.text = json.dumps(variables["media_types"])

    return [
        mock_organizations_response,
        mock_media_entries_response,
        mock_vehicle_response,
        mock_campaigns_response,
        mock_media_types_response,
    ]


@pytest.mark.basic
def test_mediatool_credentials():
    """Test Mediatool credentials."""
    MediatoolCredentials(user_id="test_user", token="test_token")  # noqa: S106


@pytest.mark.basic
@patch("viadot.sources.mediatool.get_source_credentials", return_value=None)
def test_missing_credentials(mock_get_source_credentials):
    """Test raise error without credentials."""
    with pytest.raises(CredentialError):
        Mediatool()

    mock_get_source_credentials.assert_called_once()


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_get_organizations(mock_handle_api_response):
    """Test Mediatool `_get_organizations` function."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["organizations"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._get_organizations(user_id="test_user")
    expected_result = [{"_id": "1", "name": "Org1", "abbreviation": "O1"}]
    assert result == expected_result


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_get_media_entries(mock_handle_api_response):
    """Test Mediatool `_get_media_entries` function."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["media_entries"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._get_media_entries(organization_id="org_id")
    expected_result = [
        {
            "_id": "1",
            "campaignId": "1",
            "organizationId": "1",
            "vehicleId": "1",
            "mediaTypeId": "1",
            "name": "Entry1",
            "customCol1": "CustomValue1",
            "customCol2": "CustomValue2",
        }
    ]
    assert result == expected_result


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_get_vehicles(mock_handle_api_response):
    """Test Mediatool `_get_vehicles` function."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["vehicle"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._get_vehicles(vehicle_ids=["1"])
    expected_result = [{"_id": "1", "name": "Vehicle1"}]
    assert result == expected_result


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_get_campaigns(mock_handle_api_response):
    """Test Mediatool `_get_campaigns` function."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["campaigns"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._get_campaigns(organization_id="org_id")
    expected_result = [
        {"_id": "1", "name": "Campaign1", "conventionalName": "conventionalName1"}
    ]
    assert result == expected_result


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_get_media_types(mock_handle_api_response):
    """Test Mediatool `_get_media_types` function."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["media_types"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._get_media_types(media_type_ids=["1"])
    expected_result = [{"_id": "1", "name": "Type1", "type": "Type"}]
    assert result == expected_result


@pytest.mark.connect
@patch("viadot.sources.mediatool.handle_api_response")
def test_to_records(mock_handle_api_response):
    """Test Mediatool `_to_records` method."""
    mock_response = MagicMock()
    mock_response.text = json.dumps(variables["organizations"])
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result = mediatool._to_records(endpoint="organizations")
    expected_result = [{"_id": "1", "name": "Org1", "abbreviation": "O1"}]
    assert result == expected_result


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_fetch_and_transform(mock_handle_api_response):
    """Test Mediatool `to_df` method."""
    mock_response = MagicMock()
    mock_response.text = json.dumps({"mediaEntries": [{"_id": "1", "name": "Entry1"}]})
    mock_handle_api_response.return_value = mock_response

    mediatool = Mediatool(credentials=variables["credentials"])

    result_df = mediatool.fetch_and_transform(endpoint="media_entries")

    expected_result = pd.DataFrame(
        {"_id_media_entries": ["1"], "name_media_entries": ["Entry1"]}
    )
    pd.testing.assert_frame_equal(result_df, expected_result)


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_to_df(mock_handle_api_response, mock_multiple_api_responses):
    """Test Mediatool `to_df` method."""
    mock_handle_api_response.side_effect = mock_multiple_api_responses

    mediatool = Mediatool(credentials=variables["credentials"])

    result_df = mediatool.to_df(organization_ids=["1"])
    result_df.drop(
        columns=["_viadot_source", "_viadot_downloaded_at_utc"],
        inplace=True,
        axis=1,
    )
    expected_result = pd.DataFrame(
        {
            "_id": ["1"],
            "campaignId": ["1"],
            "organizationId": ["1"],
            "vehicleId": ["1"],
            "mediaTypeId": ["1"],
            "name": ["Entry1"],
            "customCol1": "CustomValue1",
            "customCol2": "CustomValue2",
            "_id_organizations": ["1"],
            "name_organizations": ["Org1"],
            "abbreviation_organizations": ["O1"],
            "_id_campaigns": ["1"],
            "name_campaigns": ["Campaign1"],
            "conventionalName_campaigns": ["conventionalName1"],
            "_id_vehicles": ["1"],
            "name_vehicles": ["Vehicle1"],
            "_id_media_types": ["1"],
            "name_media_types": ["Type1"],
        }
    )

    pd.testing.assert_frame_equal(result_df, expected_result)


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_to_df_media_entries_columns(
    mock_handle_api_response, mock_multiple_api_responses
):
    """Test Mediatool `to_df` method, with optional media_entries_columns."""
    mock_handle_api_response.side_effect = mock_multiple_api_responses

    mediatool = Mediatool(credentials=variables["credentials"])

    result_df = mediatool.to_df(
        organization_ids=["1"],
        media_entries_columns=[
            "_id",
            "campaignId",
            "organizationId",
            "vehicleId",
            "mediaTypeId",
            "name",
        ],
    )
    result_df.drop(
        columns=["_viadot_source", "_viadot_downloaded_at_utc"],
        inplace=True,
        axis=1,
    )
    expected_result = pd.DataFrame(
        {
            "_id": "1",
            "campaignId": "1",
            "organizationId": "1",
            "vehicleId": "1",
            "mediaTypeId": "1",
            "name": "Entry1",
            "_id_organizations": ["1"],
            "name_organizations": ["Org1"],
            "abbreviation_organizations": ["O1"],
            "_id_campaigns": ["1"],
            "name_campaigns": ["Campaign1"],
            "conventionalName_campaigns": ["conventionalName1"],
            "_id_vehicles": ["1"],
            "name_vehicles": ["Vehicle1"],
            "_id_media_types": ["1"],
            "name_media_types": ["Type1"],
        }
    )

    pd.testing.assert_frame_equal(result_df, expected_result)


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_to_df_organization_ids_empty_list(mock_handle_api_response):
    """Test Mediatool `to_df` method when `organization_ids` is empty list."""
    mediatool = Mediatool(credentials=variables["credentials"])
    mock_organizations_response = MagicMock()
    mock_organizations_response.text = json.dumps(variables["organizations"])
    mock_handle_api_response.return_value = mock_organizations_response

    with pytest.raises(
        ValueError, match="'organization_ids' must be a non-empty list."
    ):
        mediatool.to_df(organization_ids=[])


@pytest.mark.functions
@patch("viadot.sources.mediatool.handle_api_response")
def test_to_df_invalid_organization_id(mock_handle_api_response):
    """Test Mediatool `to_df` method with an invalid `organization_id`."""
    mock_organizations_response = MagicMock()
    mock_organizations_response.text = json.dumps(variables["organizations"])
    mediatool = Mediatool(credentials=variables["credentials"])
    mock_handle_api_response.return_value = mock_organizations_response

    with pytest.raises(
        ValueError, match="Organization - 2 not found in organizations list."
    ):
        mediatool.to_df(organization_ids=["2"])



from viadot.config import get_source_config
from viadot.orchestration.prefect.flows import outlook_to_adls


def test_outlook_to_adls(adls_credentials_secret, outlook_config_key):
    mailbox = get_source_config("outlook").get("mailbox")
    state = outlook_to_adls(
        config_key=outlook_config_key,
        mailbox_name=mailbox,
        start_date="2023-04-12",
        end_date="2023-04-13",
        adls_azure_key_vault_secret=adls_credentials_secret,
        adls_path=f"raw/dyvenia_sandbox/genesys/{mailbox.split('@')[0].replace('.', '_').replace('-', '_')}.csv",
        adls_path_overwrite=True,
    )
    assert state.is_successful()



"""Imports."""

from .adls import adls_upload, df_to_adls
from .azure_sql import azure_sql_to_df
from .bcp import bcp
from .bigquery import bigquery_to_df
from .business_core import business_core_to_df
from .cloud_for_customers import cloud_for_customers_to_df
from .customer_gauge_to_df import customer_gauge_to_df
from .databricks import df_to_databricks
from .dbt import dbt_task
from .duckdb import duckdb_query
from .ecb import ecb_to_df
from .entraid import entraid_to_df
from .epicor import epicor_to_df
from .eurostat import eurostat_to_df
from .exchange_rates import exchange_rates_to_df
from .failed_test_email_notification import SmtpConfig, dbt_test_failure_notifier
from .genesys import genesys_to_df
from .git import clone_repo
from .hubspot import hubspot_to_df
from .informix import informix_query, informix_to_df
from .luma import luma_ingest_task
from .matomo import matomo_to_df
from .mediatool import mediatool_to_df
from .mindful import mindful_to_df
from .minio import df_to_minio
from .onestream import (
    create_batch_list_of_custom_subst_vars,
    onestream_run_data_management_seq,
    onestream_to_df,
)
from .outlook import outlook_to_df
from .postgresql import (
    create_postgresql_table,
    postgresql_query,
    postgresql_to_df,
)
from .redshift_spectrum import df_to_redshift_spectrum
from .s3 import s3_upload_file
from .salesforce import salesforce_to_df
from .sap_bw import sap_bw_to_df
from .sap_rfc import sap_rfc_to_df
from .sftp import sftp_list, sftp_to_df
from .sharepoint import (
    sharepoint_download_file,
    sharepoint_list_to_df,
    sharepoint_to_df,
)
from .sql_server import create_sql_server_table, sql_server_query, sql_server_to_df
from .supermetrics import supermetrics_to_df
from .tm1 import tm1_to_df
from .vid_club import vid_club_to_df


__all__ = [
    "SmtpConfig",
    "adls_upload",
    "azure_sql_to_df",
    "bcp",
    "bigquery_to_df",
    "business_core_to_df",
    "clone_repo",
    "cloud_for_customers_to_df",
    "create_batch_list_of_custom_subst_vars",
    "create_postgresql_table",
    "create_sql_server_table",
    "customer_gauge_to_df",
    "dbt_task",
    "dbt_test_failure_notifier",
    "df_to_adls",
    "df_to_databricks",
    "df_to_minio",
    "df_to_redshift_spectrum",
    "duckdb_query",
    "ecb_to_df",
    "entraid_to_df",
    "epicor_to_df",
    "eurostat_to_df",
    "exchange_rates_to_df",
    "genesys_to_df",
    "hubspot_to_df",
    "informix_query",
    "informix_to_df",
    "luma_ingest_task",
    "matomo_to_df",
    "mediatool_to_df",
    "mindful_to_df",
    "onestream_run_data_management_seq",
    "onestream_to_df",
    "outlook_to_df",
    "postgresql_query",
    "postgresql_to_df",
    "s3_upload_file",
    "salesforce_to_df",
    "sap_bw_to_df",
    "sap_rfc_to_df",
    "sftp_list",
    "sftp_to_df",
    "sharepoint_download_file",
    "sharepoint_list_to_df",
    "sharepoint_to_df",
    "sql_server_query",
    "sql_server_to_df",
    "supermetrics_to_df",
    "tm1_to_df",
    "vid_club_to_df",
]



"""Tests for CloudForCustomers source class."""

from datetime import datetime, timedelta

import pandas as pd

from viadot.sources.cloud_for_customers import CloudForCustomers


CONTACT_ENDPOINT = "ContactCollection"
cursor_field = "EntityLastChangedOn"
yesterday = datetime.utcnow() - timedelta(days=10)
cursor_value = yesterday.isoformat(timespec="seconds") + "Z"
cursor_filter = f"{cursor_field} ge datetimeoffset'{cursor_value}'"
filter_params = {"$filter": cursor_filter}


def test_to_df(c4c_config_key):
    c4c = CloudForCustomers(
        config_key=c4c_config_key,
        endpoint=CONTACT_ENDPOINT,
        filter_params=filter_params,
    )

    df = c4c.to_df(fields=None, dtype=None)

    assert isinstance(df, pd.DataFrame)
    assert not df.empty



"""Tasks for running query in DuckDB."""

from typing import Any, Literal

from prefect import task
from prefect.logging import get_run_logger

from viadot.config import get_source_credentials
from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import DuckDB
from viadot.sources.base import Record


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60)
def duckdb_query(
    query: str,
    fetch_type: Literal["record", "dataframe"] = "record",
    # Specifying credentials in a dictionary is not recommended in viadot tasks,
    # but in this case credentials can include only database name.
    credentials: dict[str, Any] | None = None,
    credentials_secret: str | None = None,
    config_key: str | None = None,
) -> list[Record] | bool:
    """Run query on a DuckDB database.

    Args:
        query (str, required): The query to execute on the DuckDB database.
        fetch_type (str, optional): In which form the data should be returned.
            Defaults to "record".
        credentials (dict[str, Any], optional): Credentials to the Database. Defaults to
            None.
        credentials_secret (str, optional): The name of the secret storing credentials
            to the database. Defaults to None. More info on:
            https://docs.prefect.io/concepts/blocks/
        config_key (str, optional): The key in the viadot config holding relevant
            credentials to the database. Defaults to None.
    """
    if not (credentials or credentials_secret or config_key):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    credentials = (
        credentials
        or get_source_credentials(config_key)
        or get_credentials(credentials_secret)
    )
    duckdb = DuckDB(credentials=credentials)
    result = duckdb.run_query(query=query, fetch_type=fetch_type)
    logger.info("Query has been executed successfully.")
    return result



import pytest

from viadot.sources import Supermetrics
from viadot.sources.supermetrics import SupermetricsCredentials


@pytest.fixture
def supermetrics_credentials():
    return SupermetricsCredentials(
        user="test_user",
        api_key="test_key",  # pragma: allowlist secret
    )


@pytest.fixture
def mock_get_source_credentials(
    mocker,
    supermetrics_credentials: SupermetricsCredentials,
):
    return mocker.patch(
        "viadot.config.get_source_credentials",
        return_value={
            "user": supermetrics_credentials.user,
            "api_key": supermetrics_credentials.api_key,
        },
    )


@pytest.fixture
def supermetrics(supermetrics_credentials: SupermetricsCredentials):
    return Supermetrics(
        credentials={
            "user": supermetrics_credentials.user,
            "api_key": supermetrics_credentials.api_key,
        },
        query_params={"ds_id": "GA", "query": "test_query"},
    )


def test_to_json(mocker, supermetrics: Supermetrics):
    # Mock the handle_api_response function to simulate an API response
    mock_handle_api_response = mocker.patch(
        "viadot.sources.supermetrics.handle_api_response",
    )
    mock_response = {
        "data": [["value1", "value2"]],
        "meta": {"query": {"fields": [{"field_name": "col1"}, {"field_name": "col2"}]}},
    }
    # Set the mock to return the mock response object
    mock_handle_api_response.return_value.json.return_value = mock_response

    # Call the method under test
    response = supermetrics.to_json()

    # Assert that the response is as expected
    assert response == {
        "data": [["value1", "value2"]],
        "meta": {"query": {"fields": [{"field_name": "col1"}, {"field_name": "col2"}]}},
    }


def test_to_df_with_data(supermetrics: Supermetrics, mocker):
    # Mock the API response with some data
    mock_response = {
        "meta": {
            "query": {
                "ds_id": "GA",  # Data source ID, e.g., Google Analytics
                "fields": [
                    {
                        "field_name": "date",
                        "field_type": "DIMENSION",
                        "field_split": "row",
                    },
                    {
                        "field_name": "sessions",
                        "field_type": "METRIC",
                        "field_split": "row",
                    },
                ],
                "other_query_metadata": "...",
            },
            "status": "success",  # Status of the query
            "execution_time": "0.456",  # Time taken to execute the query
        },
        "data": [
            ["2023-01-01", 100],  # Example data rows
            ["2023-01-02", 150],
            ["2023-01-03", 120],
        ],
        "paging": {
            "current_page": 1,  # Current page number if pagination is used
            "total_pages": 1,
            "total_results": 3,
        },
    }

    mock_method = mocker.patch("viadot.sources.supermetrics.Supermetrics.to_json")
    mock_method.return_value = mock_response
    mock_method = mocker.patch(
        "viadot.sources.supermetrics.Supermetrics._get_col_names",
    )
    mock_method.return_value = ["date", "sessions"]
    df = supermetrics.to_df()

    assert not df.empty
    assert list(df.columns) == [
        "date",
        "sessions",
        "_viadot_source",
        "_viadot_downloaded_at_utc",
    ]


def test_get_col_names_google_analytics_pivoted(supermetrics: Supermetrics):
    mock_response = {
        "meta": {
            "query": {
                "fields": [
                    {"field_name": "ga:date", "field_split": "column"},
                    {"field_name": "ga:sessions", "field_split": "row"},
                ],
            },
        },
        "data": [{"ga:date": "2023-01-01", "ga:sessions": 100}],
    }
    columns = supermetrics._get_col_names_google_analytics(mock_response)
    assert columns == {"ga:date": "2023-01-01", "ga:sessions": 100}


def test_get_col_names_google_analytics_non_pivoted(supermetrics: Supermetrics):
    mock_response = {
        "meta": {
            "query": {
                "fields": [
                    {"field_name": "ga:date", "field_split": "row"},
                    {"field_name": "ga:sessions", "field_split": "row"},
                ],
            },
        },
        "data": [{"ga:date": "2023-01-01", "ga:sessions": 100}],
    }
    columns = supermetrics._get_col_names_google_analytics(mock_response)
    assert columns == ["ga:date", "ga:sessions"]


def test_to_df_metadata_columns(mocker, supermetrics: Supermetrics):
    # Mock the API response with some data
    mock_response = {
        "data": [["2023-01-01", 100]],
        "meta": {
            "query": {"fields": [{"field_name": "date"}, {"field_name": "sessions"}]},
        },
    }

    mocker.patch(
        "viadot.sources.supermetrics.Supermetrics.to_json",
        return_value=mock_response,
    )
    mocker.patch(
        "viadot.sources.supermetrics.Supermetrics._get_col_names",
        return_value=["date", "sessions"],
    )

    df = supermetrics.to_df()

    assert "_viadot_source" in df.columns
    assert "_viadot_downloaded_at_utc" in df.columns


def test_get_col_names_ga(mocker, supermetrics: Supermetrics):
    mocker.patch(
        "viadot.sources.supermetrics.Supermetrics.to_json",
        return_value={
            "meta": {
                "query": {
                    "fields": [
                        {"field_name": "ga:date", "field_split": "column"},
                        {"field_name": "ga:sessions", "field_split": "row"},
                    ],
                },
            },
            "data": [{"ga:date": "2023-01-01", "ga:sessions": 100}],
        },
    )
    columns = supermetrics._get_col_names()
    assert columns == {"ga:date": "2023-01-01", "ga:sessions": 100}



"""Task for connecting to Supermetrics API."""

import pandas as pd
from prefect import task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import Supermetrics


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=60 * 60)
def supermetrics_to_df(
    query_params: dict,
    config_key: str | None = None,
    credentials_secret: str | None = None,
) -> pd.DataFrame:
    """Task to retrive data from Supermetrics and returns it as a pandas DataFrame.

    This function queries the Supermetrics API using the provided query parameters and
    returns the data as a pandas DataFrame. The function supports both
    configuration-based and secret-based credentials.

    The function is decorated with a Prefect task, allowing it to handle retries,
    logging, and timeout behavior.

    Args:
    ----
        query_params (dict):
            A dictionary containing the parameters for querying the Supermetrics API.
            These parameters define what data to retrieve and how the query should
            be constructed.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        credentials_secret (str, optional):
            The name of the secret in your secret management system that contains
            the Supermetrics API credentials. If `config_key` is not provided,
            this secret is used to authenticate with the Supermetrics API.

    Returns:
    -------
        pd.DataFrame:
            A pandas DataFrame containing the data retrieved from Supermetrics based
            on the provided query parameters.

    Raises:
    ------
        MissingSourceCredentialsError:
            Raised if neither `credentials_secret` nor `config_key` is provided,
            indicating that no valid credentials were supplied to access
            the Supermetrics API.

    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    credentials = get_credentials(credentials_secret) if not config_key else None

    supermetrics = Supermetrics(
        credentials=credentials,
        config_key=config_key,
    )
    return supermetrics.to_df(query_params=query_params)



"""Tasks for interacting with Microsoft Sharepoint."""

from typing import Any

import pandas as pd
from prefect import get_run_logger, task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import Sharepoint, SharepointList


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60)
def sharepoint_to_df(
    url: str,
    sheet_name: str | list[str | int] | int | None = None,
    columns: str | list[str] | list[int] | None = None,
    tests: dict[str, Any] | None = None,
    file_sheet_mapping: dict | None = None,
    na_values: list[str] | None = None,
    credentials_secret_basic_auth: str | None = None,
    credentials_secret_cert_auth: str | None = None,
    credentials_secret_cert_password: str | None = None,
    config_key: str | None = None,
) -> pd.DataFrame:
    """Load an Excel file stored on Microsoft Sharepoint into a pandas `DataFrame`.

    Modes:
    If the `URL` ends with the file (e.g ../file.xlsx) it downloads only the file and
    creates a DataFrame from it.
    If the `URL` ends with the folder (e.g ../folder_name/): it downloads multiple files
    and creates a DataFrame from them:
        - If `file_sheet_mapping` is provided, it downloads and processes only
            the specified files and sheets.
        - If `file_sheet_mapping` is NOT provided, it downloads and processes all of
            the files from the chosen folder.

    Args:
        url (str): The URL to the file.
        sheet_name (str | list | int, optional): Strings are used
            for sheet names. Integers are used in zero-indexed sheet positions
            (chart sheets do not count as a sheet position). Lists of strings/integers
            are used to request multiple sheets. Specify None to get all worksheets.
            Defaults to None.
        columns (str | list[str] | list[int], optional): Which columns to ingest.
            Defaults to None.
        credentials_secret_basic_auth (str, optional): The name of the secret storing
            the credentials for basic authentication. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        credentials_secret_cert_auth (str, optional): The name of the secret storing
            the credentials for certificate authentication. Defaults to None.
        credentials_secret_cert_password (str, optional): The name of the secret storing
            the password for the certificate file. Defaults to None.
        file_sheet_mapping (dict): A dictionary where keys are filenames and values are
            the sheet names to be loaded from each file. If provided, only these files
            and sheets will be downloaded. Defaults to None.
        na_values (list[str] | None): Additional strings to recognize as NA/NaN.
            If list passed, the specific NA values for each column will be recognized.
            Defaults to None.
        tests (dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from viadot.utils. Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.

    Returns:
        pd.Dataframe: The pandas `DataFrame` containing data from the file.
    """
    if not (
        credentials_secret_basic_auth or credentials_secret_cert_auth or config_key
    ):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    if credentials_secret_cert_auth and credentials_secret_cert_password:
        binary_certificate = get_credentials(secret_name=credentials_secret_cert_auth)
        credentials = get_credentials(secret_name=credentials_secret_cert_password)
        credentials["certificate"] = binary_certificate
    else:
        credentials = get_credentials(secret_name=credentials_secret_basic_auth)

    s = Sharepoint(
        credentials=credentials,
        config_key=config_key,
    )

    logger.info(f"Downloading data from {url}...")
    df = s.to_df(
        url,
        sheet_name=sheet_name,
        tests=tests,
        usecols=columns,
        na_values=na_values,
        file_sheet_mapping=file_sheet_mapping,
    )
    logger.info(f"Successfully downloaded data from {url}.")

    return df


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60)
def sharepoint_download_file(
    url: str,
    to_path: str,
    credentials_secret_basic_auth: str | None = None,
    credentials_secret_cert_auth: str | None = None,
    credentials_secret_cert_password: str | None = None,
    config_key: str | None = None,
) -> None:
    """Download a file from Sharepoint.

    Args:
        url (str): The URL of the file to be downloaded.
        to_path (str): Where to download the file.
        credentials_secret_basic_auth (str, optional): The name of the secret that
            stores Sharepoint credentials. Defaults to None.
        credentials_secret_cert_auth (str, optional): The name of the secret that stores
            Sharepoint credentials for certificate authentication. Defaults to None.
        credentials_secret_cert_password (str, optional): The name of the secret storing
            the password for the certificate file. Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials.
    """
    if not (
        credentials_secret_basic_auth or credentials_secret_cert_auth or config_key
    ):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    if credentials_secret_cert_auth and credentials_secret_cert_password:
        binary_certificate = get_credentials(secret_name=credentials_secret_cert_auth)
        credentials = get_credentials(secret_name=credentials_secret_cert_password)
        credentials["certificate"] = binary_certificate
    else:
        credentials = get_credentials(secret_name=credentials_secret_basic_auth)

    s = Sharepoint(
        credentials=credentials,
        config_key=config_key,
    )

    logger.info(f"Downloading data from {url}...")
    s.download_file(url=url, to_path=to_path)
    logger.info(f"Successfully downloaded data from {url}.")


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60)
def sharepoint_list_to_df(
    list_name: str,
    list_site: str,
    default_protocol: str | None = "https://",
    query: str | None = None,
    select: list[str] | None = None,
    credentials_secret_basic_auth: str | None = None,
    credentials_secret_cert_auth: str | None = None,
    credentials_secret_cert_password: str | None = None,
    config_key: str | None = None,
    tests: dict[str, Any] | None = None,
) -> pd.DataFrame:
    """Retrieve data from a SharePoint list into a pandas DataFrame.

    Args:
        list_name (str): The name of the SharePoint list.
        list_site (str): The Sharepoint site on which the list is stored.
        default_protocol (str, optional): The default protocol to use for
                SharePoint URLs.
                Defaults to "https://".
        query (str, optional): A query to filter items. Defaults to None.
        select (list[str], optional): Fields to include in the response.
            Defaults to None.
        credentials_secret_basic_auth (str, optional): The name of the secret storing
            the credentials. Defaults to None.
        credentials_secret_cert_auth (str, optional): The name of the secret storing
            the credentials for certificate authentication. Defaults to None.
        credentials_secret_cert_password (str, optional): The name of the secret storing
            the password for the certificate file. Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        tests (dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from viadot.utils. Defaults to None.

    Returns:
        pd.DataFrame: The DataFrame containing data from the SharePoint list.

    Raises:
        MissingSourceCredentialsError: If neither credentials_secret nor
            config_key is provided.
    """
    if not (
        credentials_secret_basic_auth or credentials_secret_cert_auth or config_key
    ):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    if credentials_secret_cert_auth and credentials_secret_cert_password:
        binary_certificate = get_credentials(secret_name=credentials_secret_cert_auth)
        credentials = get_credentials(secret_name=credentials_secret_cert_password)
        credentials["certificate"] = binary_certificate
    else:
        credentials = get_credentials(secret_name=credentials_secret_basic_auth)

    sp = SharepointList(
        credentials=credentials,
        config_key=config_key,
        default_protocol=default_protocol,
    )

    logger.info(f"Retrieving data from SharePoint list {list_name}...")
    df = sp.to_df(
        list_name=list_name,
        query=query,
        select=select,
        list_site=list_site,
        tests=tests,
    )
    logger.info(f"Successfully retrieved data from SharePoint list {list_name}.")

    return df



from loguru import logger
import pytest

from viadot.orchestration.prefect.tasks import luma_ingest_task


@pytest.mark.asyncio
async def test_luma_ingest_task_model_metadata(LUMA_URL):
    logs = await luma_ingest_task.fn(
        metadata_kind="model",
        metadata_dir_path="tests/resources/metadata/model",
        luma_url=LUMA_URL,
        logger=logger,
        raise_on_failure=False,
    )
    log = "\n".join(logs)
    success_message = "The request was successful!"

    assert success_message in log


@pytest.mark.asyncio
async def test_luma_ingest_task_model_run_metadata(LUMA_URL):
    logs = await luma_ingest_task.fn(
        metadata_kind="model_run",
        metadata_dir_path="tests/resources/metadata/model_run",
        luma_url=LUMA_URL,
        logger=logger,
        raise_on_failure=False,
    )
    log = "\n".join(logs)
    success_message = "The request was successful!"

    assert success_message in log


@pytest.mark.asyncio
async def test_luma_ingest_task_model_run_metadata_follow(LUMA_URL):
    logs = await luma_ingest_task.fn(
        metadata_kind="model_run",
        metadata_dir_path="tests/resources/metadata/model_run",
        luma_url=LUMA_URL,
        follow=True,
        logger=logger,
        raise_on_failure=False,
    )
    log = "\n".join(logs)
    success_message = "The request was successful!"

    assert success_message in log



import pytest

from viadot.config import get_source_config
from viadot.orchestration.prefect.flows import supermetrics_to_adls


@pytest.mark.parametrize(
    ("supermetrics_config_key", "adls_credentials_secret"),
    [
        ("supermetrics", "supermetrics"),
    ],
)
def test_supermetrics_to_adls(supermetrics_config_key, adls_credentials_secret):
    supermetrics_config = get_source_config(supermetrics_config_key)
    google_ads_params = {
        "ds_id": "AW",
        "ds_user": supermetrics_config["credentials"].get("user"),
        "ds_accounts": ["1007802423"],
        "date_range_type": "last_month",
        "fields": [
            "Date",
            "Campaignname",
            "Clicks",
        ],
        "max_rows": 1,
    }

    state = supermetrics_to_adls(
        query_params=google_ads_params,
        supermetrics_config_key=supermetrics_config_key,
        adls_credentials_secret=adls_credentials_secret,
        overwrite=True,
        adls_path="raw/supermetrics/.parquet",
    )

    all_successful = all(s.type == "COMPLETED" for s in state)
    assert all_successful, "Not all tasks in the flow completed successfully."



"""Common utilities for use in tasks."""

from collections.abc import Callable
import contextlib
from functools import wraps
from inspect import Parameter, iscoroutinefunction, signature
import json
from json.decoder import JSONDecodeError
import logging
import os
import re
import sys
import tempfile
from typing import Any, TypeVar

from anyio import open_process
from anyio.streams.text import TextReceiveStream
import pendulum
from prefect.blocks.core import Block
from prefect.blocks.system import Secret
from prefect.client.orchestration import get_client
from prefect.utilities.asyncutils import run_coro_as_sync
from prefect.utilities.timeout import timeout, timeout_async
from prefect_sqlalchemy import SqlAlchemyConnector


with contextlib.suppress(ModuleNotFoundError):
    from prefect_aws import AwsCredentials
    from prefect_aws.secrets_manager import AwsSecret

with contextlib.suppress(ModuleNotFoundError):
    from prefect_azure import AzureKeyVaultSecretReference

from viadot.orchestration.prefect.exceptions import MissingPrefectBlockError


T = TypeVar("T", bound=Block)
F = TypeVar("F", bound=Callable[..., object])

DEFAULT_TIMEOUT_SECONDS = 2 * 60 * 60


def with_flow_timeout_param(
    default_timeout_seconds: int = DEFAULT_TIMEOUT_SECONDS,
) -> Callable[[F], F]:
    """Add a flow-level ``timeout_seconds`` parameter to a callable.

    The decorator updates the wrapped callable's signature so Prefect can expose
    ``timeout_seconds`` as a flow parameter, then executes the callable inside a
    timeout context. The injected parameter is keyword-only and defaults to
    ``default_timeout_seconds``.

    This decorator is intended to be applied below ``@flow`` so it runs first:

        @flow(...)
        @with_flow_timeout_param()
        def my_flow(...):
            ...

    If the wrapped callable already defines ``timeout_seconds``, the original
    callable is returned unchanged.

    Args:
        default_timeout_seconds: Default timeout applied when the caller does not
            provide ``timeout_seconds`` explicitly.

    Returns:
        A decorator that preserves the wrapped callable's metadata and signature
        while adding runtime timeout enforcement.
    """

    def decorator(func: F) -> F:
        func_signature = signature(func)
        if "timeout_seconds" in func_signature.parameters:
            return func

        timeout_param = Parameter(
            "timeout_seconds",
            kind=Parameter.KEYWORD_ONLY,
            default=default_timeout_seconds,
            annotation=int,
        )

        params = list(func_signature.parameters.values())
        kwargs_idx = next(
            (idx for idx, p in enumerate(params) if p.kind == Parameter.VAR_KEYWORD),
            None,
        )
        if kwargs_idx is None:
            new_params = [*params, timeout_param]
        else:
            new_params = [*params[:kwargs_idx], timeout_param, *params[kwargs_idx:]]

        new_signature = func_signature.replace(parameters=new_params)

        if iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapped(*args: object, **kwargs: object) -> object:
                timeout_seconds = kwargs.pop("timeout_seconds", default_timeout_seconds)
                with timeout_async(seconds=timeout_seconds):
                    return await func(*args, **kwargs)

            async_wrapped.__signature__ = new_signature
            return async_wrapped  # type: ignore[return-value]

        @wraps(func)
        def wrapped(*args: object, **kwargs: object) -> object:
            timeout_seconds = kwargs.pop("timeout_seconds", default_timeout_seconds)
            with timeout(seconds=timeout_seconds):
                return func(*args, **kwargs)

        wrapped.__signature__ = new_signature
        return wrapped  # type: ignore[return-value]

    return decorator


class DynamicDateHandler:
    def __init__(
        self,
        dynamic_date_symbols: list[str] = ["<<", ">>"],  # noqa: B006
        dynamic_date_format: str = "%Y%m%d",
        dynamic_date_timezone: str = "Europe/Warsaw",
    ):
        """This class processes time-related patterns in the provided text.

        Replaces dynamic date markers with actual dates.
        The supported patterns include:
        - "today"
        - "yesterday"
        - "current_month"
        - "last_month"
        - "current_year"
        - "last_year"
        - "now_time"
        - "last_day_previous_month"
        - "last_day_of_month_year": e.g., "last_day_of_February_2020"
        - "X_years_ago_year": e.g., "3_years_ago_year" ,refers to only the year
            of the day X years ago
        - "X_years/months/days_ago_full_date":  e.g., "3_years_ago_full_date",
            refers to a given date X units ago in dynamic_date_format
        - "years_from_x_until_now_included": e.g. "years_from_2019_until_now_included",
           refers to a date range of the years from a given year
           until the current year included
        - "years_from_x_until_y_years_ago": e.g., "years_from_2019_until_y_years_ago",
          refers to a date range of the years from a given year X until y years ago
        - "last_X_years/months/days": e.g., "last_10_months", refers to a data range
            of the months in 'YMM' format
        - "first_X_days_from_X": e.g., "first_10_days_of_January_2020",
            returns a data range of days from a given month

        Args:
            dynamic_date_symbols (list[str], optional): The symbols that mark
                the start and the end of a dynamic date pattern in a text.
                Defaults to ["<<", ">>"].
            dynamic_date_format (str, optional): A date and time format string
                defining the text representation of date. Defaults to "%Y%m%d".
            dynamic_date_timezone (str, optional): A string that sets the default
                timezone used by all datetime functions. Defaults to "Europe/Warsaw".
        """
        self.singular_patterns = {
            "last_day_of_month": r"last_day_of_(\w+)_(\d{4})",
            "x_units_ago_full_date": r"(\d+)_(years?|months?|days?)_ago_full_date",
            "x_years_ago_year": r"(\d+)_years_ago_year",
        }
        self.range_patterns = {
            "last_x_units": r"last_(\d+)_(years|months|days)",
            "first_x_days_from": r"first_(\d+)_days_from_(\w+)_(\d{4})",
            "last_x_days_from": r"last_(\d+)_days_from_(\w+)_(\d{4})",
            "years_from_x_until_now": r"years_from_(\d{4})_until_now",
            "years_from_x_until_y_years_ago": r"years_from_(\d{4})_until_(\d+)_years_ago",
        }
        self.dynamic_date_format = dynamic_date_format
        self.dynamic_date_timezone = dynamic_date_timezone
        self.dynamic_date_symbols = dynamic_date_symbols
        self.replacements = self._create_date_dict()

    def _generate_years(
        self, last_years: int | None, from_year: str | None, end_year: str | None
    ) -> list[str]:
        """Generate a list of years either for the last X years or from a start year.

        Args:
            last_years (int | None): The number of years to generate
                from the current year.
            from_year (str | None): The starting year.
            end_year (int | None): The ending year.

        Returns:
            list: A list of years in ascending order.
        """
        current_year = pendulum.now().year
        if last_years:
            return [str(current_year - i) for i in range(last_years)][
                ::-1
            ]  # Reversed to ascending order
        if from_year and end_year:
            return [str(year) for year in range(int(from_year), int(end_year))]

        if from_year:
            return [str(year) for year in range(int(from_year), current_year + 1)]

        return []

    def _generate_months(self, last_months: int) -> list[str]:
        """Generate a list of first days of the last X months.

        Args:
            last_months (int): The number of months to include from the past.

        Returns:
            list: A list of dates representing the last X months in ascending order.
        """
        current_date = pendulum.now()

        return [
            current_date.subtract(months=i).start_of("month").format("YMM")
            for i in range(last_months)
        ][::-1]  # Reversed to ascending order

    def _generate_dates(self, last_days: int) -> list[str]:
        """Generate a list of dates for the last X days.

        Args:
            last_days (int): The number of days to include from the past.

        Returns:
            list: A list of dates in ascending order.
        """
        current_date = pendulum.now(self.dynamic_date_timezone)

        return [
            current_date.subtract(days=i).format("YMMDD") for i in range(last_days)
        ][::-1]  # Reversed to ascending order

    def _process_first_days(
        self, month_name: str, year: int, num_days: int
    ) -> list[str]:
        """Generate a list of the first X days of a given month and year.

        Args:
            month_name (str): The name of the month.
            year (str): The year.
            num_days (int): The number of days to include.

        Returns:
            list: A list of dates for the first X days in ascending order.
        """
        start_date = pendulum.datetime(
            int(year),
            pendulum.parse(month_name, strict=False).month,  # type: ignore
            1,  # type: ignore
        )

        return [
            start_date.add(days=i).format("YMMDD") for i in range(num_days)
        ]  # Ascending order

    def _process_last_days(
        self, month_name: str, year: int, num_days: int
    ) -> list[str]:
        """Generate a list of the last X days of a given month and year.

        Args:
            month_name (str): The name of the month.
            year (str): The year.
            num_days (int): The number of days to include.

        Returns:
            list: A list of dates for the last X days in ascending order.
        """
        start_date = pendulum.datetime(
            int(year),
            pendulum.parse(month_name, strict=False).month,  # type: ignore
            1,  # type: ignore
        )
        end_date = start_date.end_of("month")

        return [end_date.subtract(days=i).format("YMMDD") for i in range(num_days)][
            ::-1
        ]  # Reversed to ascending order

    def _process_last_day_of_month(
        self, year: str, month_name: str
    ) -> pendulum.DateTime:  # type: ignore
        """Retrieve the last day of a specified month and year.

        Args:
            year (str): The year.
            month_name (str): The name of the month.

        Returns:
            pendulum.DateTime: A date object containing the last day of the given month.
        """
        month_num = pendulum.parse(month_name, strict=False).month  # type: ignore

        return pendulum.datetime(int(year), month_num, 1).end_of("month")

    def _process_x_years_ago(self, year: int) -> str:
        """Retrieve the year of a date X years from now.

        Args:
            year (int): The year.

        Returns:
            str: A string containing the year of the specified time ago.
        """
        current_date = pendulum.now()

        return current_date.subtract(years=year).format("Y")

    def _get_date_x_ago_full_date(
        self, number: int, unit: str
    ) -> pendulum.DateTime | None:  # type: ignore
        """Retrieve the full date for X units ago from today.

        Args:
            number (int): The number of units (years, months, days).
            unit (str): The unit of time ('years', 'months', 'days').

        Returns:
            pendulum.DateTime: A date for X units ago from today.
        """
        return {
            "years": pendulum.now(self.dynamic_date_timezone).subtract(years=number),
            "months": pendulum.now(self.dynamic_date_timezone).subtract(months=number),
            "days": pendulum.now(self.dynamic_date_timezone).subtract(days=number),
        }.get(unit)

    def _create_date_dict(self) -> dict[str, str]:
        """Create and return a key phrase: dynamic date value dictionary.

        Dictionary values "today", "yesterday" and "last_year_previous_month" are
        formatted into the dynamic_date_format.

        The other values and their formatting:
            - "current_month" - A string date formatted with a string format '%m'.
            - "last_month" - A string date formatted with a format "%mm".
            - "current_year" - A string date formatted with a format "%mm".
            - "last_year" - A string date formatted with a string format '%Y'
            - "now_time" - A string date formatted with a string format '%H%M%S'.

        Returns:
            dict[str, str]: A dictionary with key phrases as keys
            and dynamically created dates as values.
        """
        today = pendulum.today(self.dynamic_date_timezone)
        yesterday = pendulum.yesterday(self.dynamic_date_timezone)
        last_month = today.subtract(months=1).month
        last_year = today.subtract(years=1)
        now_time = pendulum.now(self.dynamic_date_timezone)
        last_day_prev_month = today.subtract(months=1).end_of("month")

        return {
            "today": today.strftime(self.dynamic_date_format),
            "yesterday": yesterday.strftime(self.dynamic_date_format),
            "current_month": today.strftime("%m"),
            "last_month": f"{last_month:02d}",
            "current_year": today.strftime("%Y"),
            "last_year": last_year.strftime("%Y"),
            "now_time": now_time.strftime("%H%M%S"),
            "last_day_previous_month": last_day_prev_month.strftime(
                self.dynamic_date_format
            ),
        }

    def _handle_singular_dates(
        self, dynamic_date_marker: str, match: list[tuple], key: str
    ) -> pendulum.DateTime | str | None:  # type: ignore
        """Directs execution of a specific function based on the value of `key`.

        Args:
            match (list[tuple]): list of every pattern match that occurs in a string.
            key (str): Key phrase that determines the execution of a specific function.

        Returns:
            str or pendulum.DateTime:
            - If key == 'x_years_ago_year' returns string of a pendulum date formatted
                with a pendulum token 'Y'.
            - If key != 'x_years_ago_year' returns a pendulum.DateTime
        """
        if key == "last_day_of_month":
            for month_name, year in match:
                replacement = self._process_last_day_of_month(year, month_name)

        elif key == "x_units_ago_full_date":
            for x, units in match:
                replacement = self._get_date_x_ago_full_date(int(x), units)

        elif key == "x_years_ago_year":
            for x in match:
                replacement = self._process_x_years_ago(int(x))  # type: ignore
        else:
            replacement = dynamic_date_marker
        return replacement

    def _generate_dates_based_on_unit(
        self, dynamic_date_marker: str, number: int, unit: str
    ) -> list[str] | str:
        """Direct execution of a specific function based on the value of `unit`.

        Possible values of `unit` correspond to different date formatting styles:
            - 'years': Return a date formatted with a pendulum token 'Y'.
            - 'months': Return a date formatted with a pendulum token 'YMM'.
            - 'days': Return a date with a pendulum token 'YMMDD'.

        Args:
            number (int): The number of units from the current year to include.
            unit (str): The unit of time ('years', 'months', 'days').

        Returns:
            list[str] or string:
                - If the key argument matches the given options, returns
                    a list of dynamically created dates generated based on `unit`
                    in an ascending order.
                - If the key argument doesn't match the given options,
                    returns the unhandled dynamic_date_marker parameter
        """
        if unit == "years":
            return self._generate_years(
                last_years=number, from_year=None, end_year=None
            )
        if unit == "months":
            return self._generate_months(last_months=number)
        if unit == "days":
            return self._generate_dates(last_days=number)
        return dynamic_date_marker

    def _handle_data_ranges(  # noqa: C901
        self, dynamic_date_marker: str, match_found: list[tuple], key: str
    ) -> list[str] | str:
        """Direct execution of a specific function based on the provided value of `key`.

        Depending on a unit ('years'/'months'/'days') the `match_found` refers to,
        date formatting style differs:
            - 'years': Return a date formatted with a pendulum token 'Y'.
            - 'months': Return a date formatted with a pendulum token 'YMM'.
            - 'days': Return a date with a pendulum token 'YMMDD'.

        Args:
            dynamic_date_marker (str): A dynamic date marker that has been found in text
                including the dynamic_date_symbols.
            match_found (list[tuple]): list of every pattern match that occurs in
                a string.
            key (str): Key phrase that determines the execution of a specific function.

        Returns:
            list[str] or string:
                - If the key argument matches the given options,
                    returns list of extracted date ranges in ascending order.
                - If the key argument doesn't match the given options,
                    returns the unhandled dynamic_date_marker parameter
        """
        if key == "last_x_units":
            for number, unit in match_found:
                return self._generate_dates_based_on_unit(
                    dynamic_date_marker, int(number), unit
                )

        elif key == "years_from_x_until_now":
            for start_year in match_found:
                return self._generate_years(
                    last_years=None,
                    from_year=start_year,  # type: ignore
                    end_year=None,  # type: ignore
                )
        elif key == "years_from_x_until_y_years_ago":
            for start_year, end_year in match_found:
                end_year = str(  # noqa: PLW2901
                    int(self.replacements["current_year"]) - int(end_year) + 1
                )
                return self._generate_years(
                    last_years=None,
                    from_year=start_year,  # type: ignore
                    end_year=end_year,  # type: ignore
                )
        elif key == "first_x_days_from":
            for num_days, month_name, year in match_found:
                return self._process_first_days(month_name, year, int(num_days))

        elif key == "last_x_days_from":
            for num_days, month_name, year in match_found:
                return self._process_last_days(month_name, year, int(num_days))

        return dynamic_date_marker

    def _find_dynamic_date_patterns(self, text: str) -> list[str]:
        """Retrieve the matches between the dynamic date symbols from string.

        Args:
            text (str): A string containing dynamic date patterns
                to be listed.

        Returns:
            list: A list of matches found between the dynamic date symbols.
        """
        start_symbol, end_symbol = self.dynamic_date_symbols
        start, end = re.escape(start_symbol), re.escape(end_symbol)
        pattern = rf"{start}.*?{end}"

        return re.findall(pattern, text, re.IGNORECASE)

    def _process_string(self, text: str) -> list[str] | str:
        """Analyze and extract date ranges or singular dates from the given text.

        It bases on specific patterns or pendulum dates.

        Args:
            text (str): A string containing various time-related patterns
                to be analyzed.

        Returns:
            list or string:
                - If the input is a key phrase for a data range,
                    returns list of extracted date ranges in ascending order.
                - If the input is a key phrase for a single date or a pendulum date,
                    returns the input text with an accurate date.
        """
        start_symbol, end_symbol = self.dynamic_date_symbols
        start, end = re.escape(start_symbol), re.escape(end_symbol)
        pattern = rf"{start}.*?{end}"

        matches_between_symbols = re.findall(pattern, text, re.IGNORECASE)
        if not matches_between_symbols:
            return text

        for match in matches_between_symbols:
            match_no_symbols = match[len(start_symbol) : -len(end_symbol)]
            replacement = None
            # Processing the singular dates
            for key, pattern in self.singular_patterns.items():
                match_found = re.findall(pattern, match_no_symbols, re.IGNORECASE)
                if match_found:
                    replacement = self._handle_singular_dates(match, match_found, key)

            # Process range date matches
            for key, pattern in self.range_patterns.items():
                match_found = re.findall(pattern, match_no_symbols, re.IGNORECASE)
                if match_found:
                    return self._handle_data_ranges(match, match_found, key)

            if match_no_symbols in self.replacements:
                replacement = self.replacements[match_no_symbols]
            if not replacement and bool(
                re.match(r"^\s*pendulum\.\w+\(.*\)\s*$", match_no_symbols)
            ):
                expr = re.sub(
                    r"pendulum\.(today|now)\(\)",  # pattern
                    rf'pendulum.\1("{self.dynamic_date_timezone}")',  # replacement
                    match_no_symbols,
                )
                replacement = eval(expr, {"pendulum": pendulum})  # noqa: S307
            text = text.replace(
                match,
                (
                    replacement.strftime(self.dynamic_date_format)
                    if isinstance(replacement, pendulum.DateTime)  # type: ignore
                    else replacement
                ),
            )

        return text

    def process_dates(self, processed_input: str | list[str]) -> str | list[str]:
        """Process an input by processing dates within it.

        If the input is a string, it applies the _process_string() function.
        If the input is a list, it recursively processes each list item.

        Args:
            processed_input (str or list): The segment to be processed.

        Returns:
            str or list:
                - If processed_input is a string, returns the processed string.
                - If processed_input is a list, returns a list of processed strings.
        """
        if isinstance(processed_input, str):
            return self._process_string(processed_input)
        if isinstance(processed_input, list):
            return [self.process_dates(sub_segment) for sub_segment in processed_input]  # type: ignore
        return processed_input


async def list_block_documents() -> list[Any]:
    """Retrieve list of Prefect block documents."""
    async with get_client() as client:
        return await client.read_block_documents()


def _load_prefect_block(block_cls: type[T], block_name: str) -> T:
    """Load a Prefect block in sync mode."""
    return block_cls.load(block_name, _sync=True)


def _get_azure_credentials(secret_name: str) -> dict[str, Any]:
    """Retrieve credentials from the Prefect 'AzureKeyVaultSecretReference' block.

    Args:
        secret_name (str): The name of the secret to be retrieved.

    Returns:
        dict: A dictionary containing the credentials.
    """
    azure_secret_ref = _load_prefect_block(AzureKeyVaultSecretReference, secret_name)
    try:
        credentials = json.loads(azure_secret_ref.get_secret())
    except (JSONDecodeError, TypeError):
        credentials = azure_secret_ref.get_secret()

    return credentials


def _get_aws_credentials(
    secret_name: str,
    block_type: str,
) -> dict[str, Any] | str:
    """Retrieve credentials from the Prefect 'AwsSecret' block document.

        It distinguishes the types of block to take a correct action.
        This is needed as AwsSecret block stores credentials that can be used in tasks
        which ingest data from different systems (for ex. SAP), where
        AwsCredentials block stores AWS credentials solely and this could
        be needed in other tasks which connects
        to AWS directly ( for ex. redshift spectrum , s3).

    Args:
        secret_name (str): The name of the secret to be retrieved.
        block_type (str): Type of prefect block

    Returns:
        dict | str: A dictionary or a string containing the credentials.
    """
    if block_type == "AwsSecret":
        aws_secret_block = _load_prefect_block(AwsSecret, secret_name)
        secret = aws_secret_block.read_secret()
        try:
            credentials = json.loads(secret)
        except (json.JSONDecodeError, UnicodeDecodeError, TypeError):
            credentials = secret
    elif block_type == "AwsCredentials":
        aws_credentials_block = _load_prefect_block(AwsCredentials, secret_name)
        credentials = {
            "aws_access_key_id": aws_credentials_block.aws_access_key_id,
            "aws_secret_access_key": aws_credentials_block.aws_secret_access_key.get_secret_value(),
            "region_name": aws_credentials_block.region_name,
        }

    return credentials


def _get_secret_credentials(secret_name: str) -> dict[str, Any] | str:
    """Retrieve credentials from the Prefect 'Secret' block document.

    Args:
        secret_name (str): The name of the secret to be retrieved.

    Returns:
        dict | str: A dictionary or a string containing the credentials.
    """
    secret = _load_prefect_block(Secret, secret_name).get()
    try:
        credentials = json.loads(secret)
    except (json.JSONDecodeError, TypeError):
        credentials = secret

    return credentials


def _get_database_credentials(secret_name: str) -> dict[str, Any] | str:
    """Retrieve credentials from the Prefect 'SqlAlchemyConnectord' block document.

    Args:
        secret_name (str): The name of the secret to be retrieved.

    Returns:
        dict | str: A dictionary or a string containing the credentials.
    """
    secret = _load_prefect_block(SqlAlchemyConnector, secret_name).dict()

    # Extract from connection_info structure
    conn_info = secret.get("connection_info", {})

    # connection_info might be a dict or ConnectionComponents object
    if hasattr(conn_info, "dict"):
        conn_info = conn_info.dict()

    credentials = {}
    credentials["user"] = conn_info.get("username")
    credentials["db_name"] = conn_info.get("database")

    password_obj = conn_info.get("password")
    if password_obj:
        credentials["password"] = (
            password_obj.get_secret_value()
            if hasattr(password_obj, "get_secret_value")
            else password_obj
        )
    else:
        credentials["password"] = None

    host = conn_info.get("host")
    port = conn_info.get("port")
    if port:
        credentials["server"] = f"{host},{port}"
    else:
        credentials["server"] = host

    return credentials


def get_credentials(secret_name: str) -> dict[str, Any]:
    """Retrieve credentials from the Prefect block document.

    Args:
        secret_name (str): The name of the secret to be retrieved.

    Returns:
        dict: A dictionary containing the credentials.
    """
    # Prefect does not allow upper case letters for blocks,
    # so some names might be lowercased versions of the original

    secret_name_lowercase = secret_name.lower()
    blocks = run_coro_as_sync(list_block_documents())

    for block in blocks:
        if block.name == secret_name_lowercase:
            block_type = block.block_schema.fields["title"]
            break
    else:
        msg = "The provided secret name is not valid."
        raise MissingPrefectBlockError(msg)

    if block_type in ("AwsSecret", "AwsCredentials"):
        credentials = _get_aws_credentials(secret_name, block_type)
    elif block_type == "AzureKeyVaultSecretReference":
        credentials = _get_azure_credentials(secret_name)
    elif block_type == "SqlAlchemyConnector":
        credentials = _get_database_credentials(secret_name)
    elif block_type == "Secret":
        credentials = _get_secret_credentials(secret_name)
    else:
        msg = f"The provided secret block type: {block_type} is not supported"
        raise MissingPrefectBlockError(msg)

    return credentials


async def shell_run_command(
    command: str,
    env: dict[str, Any] | None = None,
    helper_command: str | None = None,
    shell: str = "bash",
    return_all: bool = False,
    stream_level: int = logging.INFO,
    logger: logging.Logger | None = None,
    raise_on_failure: bool = True,
) -> list[str] | str:
    """Runs arbitrary shell commands as a util.

    Args:
        command: Shell command to be executed; can also be
            provided post-initialization by calling this task instance.
        env: dictionary of environment variables to use for
            the subprocess; can also be provided at runtime.
        helper_command: String representing a shell command, which
            will be executed prior to the `command` in the same process.
            Can be used to change directories, define helper functions, etc.
            for different commands in a flow.
        shell: Shell to run the command with; defaults to "bash".
        return_all: Whether this task should return all lines of stdout as a list,
            or just the last line as a string; defaults to `False`.
        stream_level: The logging level of the stream.
        logger: Can pass a desired logger; if not passed, will automatically
            gets a run logger from Prefect.
        raise_on_failure: Whether to raise an exception if the command fails.

    Returns:
        If return all, returns all lines as a list; else the last line as a string.

    Example:
        Echo "hey it works".
        ```python
        from prefect_shell.utils import shell_run_command
        await shell_run_command("echo hey it works")
        ```
    """
    if logger is None:
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger("prefect_shell.utils")

    current_env = os.environ.copy()
    current_env.update(env or {})

    with tempfile.NamedTemporaryFile(prefix="prefect-") as tmp:
        if helper_command:
            tmp.write(helper_command.encode())
            tmp.write(os.linesep.encode())
        tmp.write(command.encode())
        tmp.flush()

        shell_command = [shell, tmp.name]
        if sys.platform == "win32":
            shell_command = " ".join(shell_command)

        lines = []
        async with await open_process(shell_command, env=env) as process:
            async for text in TextReceiveStream(process.stdout):
                logger.log(stream_level, text)
                lines.extend(text.rstrip().split("\n"))

            await process.wait()
            if process.returncode:
                stderr = "\n".join(
                    [text async for text in TextReceiveStream(process.stderr)]
                )
                if not stderr and lines:
                    stderr = f"{lines[-1]}\n"
                msg = f"Command failed with exit code {process.returncode}:\n{stderr}"
                if raise_on_failure:
                    raise RuntimeError(msg)
                lines.append(msg)

    return lines if return_all else lines[-1]



import os
from pathlib import Path

import pandas as pd
from prefect import flow

from viadot.orchestration.prefect.tasks import (
    sharepoint_download_file,
    sharepoint_to_df,
)


DF1 = pd.DataFrame({"a": [1, 2], "b": [3, 4]})


def test_to_df(sharepoint_url, sharepoint_config_key):
    @flow
    def test_to_df_flow():
        return sharepoint_to_df(url=sharepoint_url, config_key=sharepoint_config_key)

    received_df = test_to_df_flow()
    assert not received_df.empty


def test_download_file(sharepoint_url, sharepoint_config_key):
    file = "sharepoint_test" + sharepoint_url.split(".")[-1]

    @flow
    def test_download_file_flow():
        return sharepoint_download_file(
            url=sharepoint_url, to_path=file, config_key=sharepoint_config_key
        )

    test_download_file_flow()

    assert file in os.listdir()

    Path(file).unlink()



"""PostgreSQL source connector."""

import logging

from pydantic import BaseModel, SecretStr
import pyodbc

from viadot.config import get_source_credentials
from viadot.sources.base import SQL


logger = logging.getLogger(__name__)


class PostgreSQLCredentials(BaseModel):
    username: str
    password: str | SecretStr | None = None


class PostgreSQL(SQL):
    def __init__(
        self,
        credentials: PostgreSQLCredentials | None = None,
        host: str = "localhost",
        port: int = 5432,
        db_name: str = "postgres",
        sslmode: str = "require",
        config_key: str | None = None,
        driver: str = "PostgreSQL Unicode",
        query_timeout: int = 60,
        *args,
        **kwargs,
    ):
        """PostgreSQL connector.

        Args:
            driver (str | None, optional): ODBC driver. Default "PostgreSQL Unicode".
            query_timeout (int, optional): Query timeout in seconds. Defaults to 60.
            host (str, optional): The host of the PostgreSQL database.
                Defaults to "localhost".
            port (int, optional): The port of the PostgreSQL database.
                Defaults to 5432.
            db_name (str, optional): The name of the PostgreSQL database.
                Defaults to "postgres".
            sslmode (str, optional): The SSL mode to use for the
                PostgreSQL database. Defaults to "require".
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to None.
            driver (str, optional): ODBC driver. Default "PostgreSQL Unicode".
            query_timeout (int, optional): Query timeout in seconds. Defaults to 60.
        """
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = PostgreSQLCredentials(**raw_creds).dict(
            by_alias=True
        )  # validate the credentials

        super().__init__(
            *args,
            credentials=validated_creds,
            driver=driver,
            query_timeout=query_timeout,
            **kwargs,
        )
        self.server = host
        self.port = port
        self.db_name = db_name
        self.user = self.credentials.get("username")
        self.password = self.credentials.get("password")
        self.sslmode = sslmode

    @property
    def conn_str(self) -> str:
        """Generate a PostgreSQL ODBC connection string from credentials.

        Returns:
            str: The ODBC connection string.
        """
        driver = self.credentials["driver"]
        server = self.server
        port = self.port
        db_name = self.db_name
        uid = self.credentials.get("username") or ""
        pwd = self.credentials.get("password") or ""

        conn_str = (
            f"DRIVER={{{driver}}};"
            f"SERVER={server};"
            f"PORT={port};"
            f"DATABASE={db_name};"
            f"UID={uid};"
            f"PWD={pwd};"
        )

        # Optional SSL mode if provided, e.g. 'require', 'verify-ca', 'disable'
        conn_str += f"SSLmode={self.sslmode};"

        return conn_str

    def _check_if_table_exists(self, table: str, schema: str | None = None) -> bool:
        """Check if a table exists in the given schema (PostgreSQL).

        Args:
            table (str): Table name.
            schema (str, optional): Schema name. Defaults to 'public' if not provided.
        """
        schema = schema or "public"
        exists_query = (
            f"SELECT 1 FROM information_schema.tables "  # noqa: S608
            f"WHERE table_schema = '{schema}' AND table_name = '{table}' LIMIT 1"
        )
        return bool(self.run(exists_query))

    @property
    def con(self) -> pyodbc.Connection:
        """Create a connection without forcing unsupported connection attributes.

        Some ODBC drivers (e.g., psqlODBC) do not support setting the connection
        timeout attribute via `self._con.timeout`. We ignore that if raised.

        Returns:
            pyodbc.Connection: The database connection.
        """
        if not self._con:
            # Keep a short login timeout to fail fast on bad hosts/ports
            self._con = __import__("pyodbc").connect(self.conn_str, timeout=5)
            try:
                # Best-effort; ignore if the driver doesn't support it
                self._con.timeout = self.query_timeout
            except Exception:
                logger.warning(
                    "The driver does not support setting the connection timeout attribute."
                )
        return self._con



"""Source for connecting to TM1."""

from typing import Any, Literal

import pandas as pd
from pydantic import BaseModel, SecretStr
from TM1py.Services import TM1Service

from viadot.config import get_source_credentials
from viadot.exceptions import ValidationError
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns


class TM1Credentials(BaseModel):
    """TM1 credentials.

    Uses simple authentication:
        - username: The user name to use for the connection.
        - password: The password to use for the connection.
        - address: The ip address to use for the connection.
        - port: The port to use for the connection.
    """

    username: str
    password: SecretStr
    address: str
    port: str


class TM1(Source):
    """TM1 connector."""

    def __init__(
        self,
        credentials: dict[str, Any] | None = None,
        config_key: str = "TM1",
        limit: int | None = None,
        private: bool = False,
        verify: bool = False,
        *args,
        **kwargs,
    ):
        """Create a TM1 connector instance.

        Args:
            credentials (dict[str, Any], optional): Credentials stored in a dictionary.
                Required credentials: username, password, address, port.
                Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to "TM1".
            mdx_query (str, optional): MDX select query needed to download the data.
                Defaults to None.
            cube (str, optional): Cube name from which data will be downloaded.
                Defaults to None.
            view (str, optional): View name from which data will be downloaded.
                Defaults to None.
            dimension (str, optional): Dimension name. Defaults to None.
            hierarchy (str, optional): Hierarchy name. Defaults to None.
            limit (str, optional): How many rows should be extracted.
                If None all the available rows will be downloaded. Defaults to None.
            private (bool, optional): Whether or not data download should be private.
                Defaults to False.
            verify (bool, optional): Whether or not verify SSL certificates.
                Defaults to False.

        Raises:
            CredentialError: When credentials are not found.

        """
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(TM1Credentials(**raw_creds))

        self.config_key = config_key
        self.limit = limit
        self.private = private
        self.verify = verify

        super().__init__(*args, credentials=validated_creds, **kwargs)

    def get_connection(self) -> TM1Service:
        """Start a connection to TM1 instance.

        Returns:
            TM1Service: Service instance if connection is successful.
        """
        return TM1Service(
            address=self.credentials.get("address"),
            port=self.credentials.get("port"),
            user=self.credentials.get("username"),
            password=self.credentials.get("password").get_secret_value(),
            ssl=self.verify,
        )

    def get_cubes_names(self) -> list:
        """Get list of available cubes in TM1 instance.

        Returns:
            list: List containing available cubes names.

        """
        conn = self.get_connection()
        return conn.cubes.get_all_names()

    def get_views_names(self, cube: str) -> list:
        """Get list of available views in TM1 cube instance.

        Args:
            cube (str): Cube name.

        Returns:
            list: List containing available views names.

        """
        conn = self.get_connection()
        return conn.views.get_all_names(cube)

    def get_dimensions_names(self) -> list:
        """Get list of available dimensions in TM1 instance.

        Returns:
            list: List containing available dimensions names.

        """
        conn = self.get_connection()
        return conn.dimensions.get_all_names()

    def get_hierarchies_names(self, dimension: str) -> list:
        """Get list of available hierarchies in TM1 dimension instance.

        Args:
            dimension (str): Dimension name.

        Returns:
            list: List containing available hierarchies names.

        """
        conn = self.get_connection()
        return conn.hierarchies.get_all_names(dimension)

    def get_available_elements(self, dimension: str, hierarchy: str) -> list:
        """Get list of available elements in TM1 instance.

        Elements are extracted based on hierarchy and dimension.

        Args:
            dimension (str): Dimension name.
            hierarchy (str): Hierarchy name.

        Returns:
            list: List containing available elements names.

        """
        if dimension is None or hierarchy is None:
            msg = "Missing dimension or hierarchy."
            raise ValidationError(msg)

        conn = self.get_connection()
        return conn.elements.get_element_names(
            dimension_name=dimension, hierarchy_name=hierarchy
        )

    @add_viadot_metadata_columns
    def to_df(
        self,
        if_empty: Literal["warn", "fail", "skip"] = "skip",
        mdx_query: str | None = None,
        cube: str | None = None,
        view: str | None = None,
    ) -> pd.DataFrame:
        """Download data into a pandas DataFrame.

            To download the data to the dataframe user needs to specify MDX query
                or combination of cube and view.

        Args:
            if_empty (Literal["warn", "fail", "skip"], optional): What to do if output
                DataFrame is empty. Defaults to "skip".
            mdx_query (str): MDX query. Defaults to None.
            cube (str): Cube name. Defaults to None.
            view (str): View name. Defaults to None.

        Returns:
            pd.DataFrame: DataFrame with the data.

        Raises:
            ValidationError: When mdx and cube + view are not specified
                or when combination of both is specified.
        """
        conn = self.get_connection()

        if mdx_query is None and (cube is None or view is None):
            error_msg = "MDX query or cube and view are required."
            raise ValidationError(error_msg)
        if mdx_query is not None and (cube is not None or view is not None):
            error_msg = "Specify only one: MDX query or cube and view."
            raise ValidationError(error_msg)
        if cube is not None and view is not None:
            df = conn.cubes.cells.execute_view_dataframe(
                cube_name=cube,
                view_name=view,
                private=self.private,
                top=self.limit,
            )
        elif mdx_query is not None:
            df = conn.cubes.cells.execute_mdx_dataframe(mdx_query)

        self.logger.info(
            f"Data was successfully transformed into DataFrame: {len(df.columns)} columns and {len(df)} rows."
        )
        if df.empty:
            self._handle_if_empty(if_empty)
        return df



"""Test ECB source."""

from pathlib import Path
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
import requests

from viadot.sources.ecb import ECBExchangeRates


@pytest.fixture
def sample_ecb_response():
    """Load sample ECB API response for testing."""
    response_path = Path(__file__).parent / "test_ecb_response.xml"
    with response_path.open(encoding="utf-8") as f:
        return f.read()


@pytest.fixture
def ecb_instance():
    """Create ECBExchangeRates instance."""
    return ECBExchangeRates()


def test_init(ecb_instance):
    """Test ECBExchangeRates initialization."""
    assert ecb_instance.credentials is None
    assert (
        ecb_instance.URL
        == "https://www.ecb.europa.eu/stats/eurofxref/eurofxref-hist.xml"
    )


@patch("viadot.sources.ecb.handle_api_response")
def test_fetch_data_success(
    mock_handle_api_response, ecb_instance, sample_ecb_response
):
    """Test successful data fetching."""
    # Setup mock response
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = sample_ecb_response
    mock_handle_api_response.return_value = mock_response

    # Call fetch_data (no url parameter needed)
    data = ecb_instance.fetch_data()

    # Verify the data was returned
    assert data == sample_ecb_response

    # Verify handle_api_response was called correctly with hardcoded URL
    mock_handle_api_response.assert_called_once()
    call_args = mock_handle_api_response.call_args
    assert call_args[1]["url"] == ecb_instance.URL
    assert call_args[1]["params"] is None
    assert call_args[1]["method"] == "GET"


@patch("viadot.sources.ecb.handle_api_response")
def test_fetch_data_request_exception(mock_handle_api_response, ecb_instance):
    """Test fetch_data with request exception."""
    from viadot.exceptions import APIError

    mock_handle_api_response.side_effect = requests.exceptions.RequestException(
        "Connection failed"
    )

    with pytest.raises(APIError):
        ecb_instance.fetch_data()


def test_parse_xml(ecb_instance, sample_ecb_response):
    """Test XML parsing."""
    df = ecb_instance._parse_xml(sample_ecb_response)

    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert "time" in df.columns
    assert "currency" in df.columns
    assert "rate" in df.columns

    # Check that we have data
    assert len(df) > 0

    # Check that time is consistent
    assert df["time"].nunique() == 1
    assert str(df["time"][0].date()) == "2025-12-30"

    # Check that currency values are strings
    assert df["currency"].dtype == "object"

    # Check that rate values are numeric
    assert pd.api.types.is_numeric_dtype(df["rate"])


def test_parse_xml_invalid_xml(ecb_instance):
    """Test parse_xml with invalid XML."""
    # Use XML that doesn't match ECB structure
    # This will parse successfully but fail structure validation
    invalid_xml = "<invalid>xml</invalid>"
    with pytest.raises(ValueError, match="Could not find time attribute"):
        ecb_instance._parse_xml(invalid_xml)


def test_parse_xml_missing_time(ecb_instance):
    """Test parse_xml with XML missing time attribute."""
    xml_without_time = """<?xml version="1.0" encoding="UTF-8"?>
    <gesmes:Envelope xmlns:gesmes="http://www.gesmes.org/xml/2002-08-01" xmlns="http://www.ecb.int/vocabulary/2002-08-01/eurofxref">
        <Cube>
            <Cube currency='USD' rate='1.1757'/>
        </Cube>
    </gesmes:Envelope>"""
    with pytest.raises(ValueError, match="Could not find time attribute"):
        ecb_instance._parse_xml(xml_without_time)


def test_parse_xml_empty_data(ecb_instance):
    """Test parse_xml with XML containing no exchange rates."""
    xml_empty = """<?xml version="1.0" encoding="UTF-8"?>
    <gesmes:Envelope xmlns:gesmes="http://www.gesmes.org/xml/2002-08-01" xmlns="http://www.ecb.int/vocabulary/2002-08-01/eurofxref">
        <Cube>
            <Cube time='2025-12-30'>
            </Cube>
        </Cube>
    </gesmes:Envelope>"""
    df = ecb_instance._parse_xml(xml_empty)
    assert df.empty
    assert list(df.columns) == ["time", "currency", "rate"]


@patch("viadot.sources.ecb.handle_api_response")
def test_to_df_success(mock_handle_api_response, ecb_instance, sample_ecb_response):
    """Test successful to_df conversion."""
    # Setup mock response
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = sample_ecb_response
    mock_handle_api_response.return_value = mock_response

    # Call to_df (no url parameter needed)
    df = ecb_instance.to_df()

    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert "time" in df.columns
    assert "currency" in df.columns
    assert "rate" in df.columns


@patch("viadot.sources.ecb.handle_api_response")
def test_to_df_empty_result_warn(mock_handle_api_response, ecb_instance, caplog):
    """Test to_df with empty result and warn if_empty."""
    xml_empty = """<?xml version="1.0" encoding="UTF-8"?>
    <gesmes:Envelope xmlns:gesmes="http://www.gesmes.org/xml/2002-08-01" xmlns="http://www.ecb.int/vocabulary/2002-08-01/eurofxref">
        <Cube>
            <Cube time='2025-12-30'>
            </Cube>
        </Cube>
    </gesmes:Envelope>"""

    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = xml_empty
    mock_handle_api_response.return_value = mock_response

    with caplog.at_level("WARNING"):
        df = ecb_instance.to_df(if_empty="warn")

    assert df.empty

    assert "No exchange rate data found in XML response." in caplog.text
    assert "No exchange rates found in the response." in caplog.text
    assert "The query produced no data." in caplog.text


@patch("viadot.sources.ecb.handle_api_response")
def test_to_df_empty_result_skip(mock_handle_api_response, ecb_instance):
    """Test to_df with empty result and skip if_empty."""
    from viadot.signals import SKIP

    xml_empty = """<?xml version="1.0" encoding="UTF-8"?>
    <gesmes:Envelope xmlns:gesmes="http://www.gesmes.org/xml/2002-08-01" xmlns="http://www.ecb.int/vocabulary/2002-08-01/eurofxref">
        <Cube>
            <Cube time='2025-12-30'>
            </Cube>
        </Cube>
    </gesmes:Envelope>"""

    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = xml_empty
    mock_handle_api_response.return_value = mock_response

    with pytest.raises(SKIP):
        ecb_instance.to_df(if_empty="skip")


@patch("viadot.sources.ecb.handle_api_response")
def test_to_df_empty_result_fail(mock_handle_api_response, ecb_instance):
    """Test to_df with empty result and fail if_empty."""
    xml_empty = """<?xml version="1.0" encoding="UTF-8"?>
    <gesmes:Envelope xmlns:gesmes="http://www.gesmes.org/xml/2002-08-01" xmlns="http://www.ecb.int/vocabulary/2002-08-01/eurofxref">
        <Cube>
            <Cube time='2025-12-30'>
            </Cube>
        </Cube>
    </gesmes:Envelope>"""

    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = xml_empty
    mock_handle_api_response.return_value = mock_response

    with pytest.raises(ValueError, match="The query produced no data"):
        ecb_instance.to_df(if_empty="fail")


@patch("viadot.sources.ecb.handle_api_response")
def test_to_df_with_tests_validation(
    mock_handle_api_response, ecb_instance, sample_ecb_response
):
    """Test to_df with tests validation."""
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = sample_ecb_response
    mock_handle_api_response.return_value = mock_response

    test_config = {"column_tests": {"currency": {"not_null": True}}}

    with patch("viadot.sources.ecb.validate") as mock_validate:
        df = ecb_instance.to_df(tests=test_config)

        mock_validate.assert_called_once()
        args, kwargs = mock_validate.call_args
        assert kwargs["df"] is df
        assert kwargs["tests"] == test_config


@patch("viadot.sources.ecb.handle_api_response")
def test_full_workflow(mock_handle_api_response, ecb_instance, sample_ecb_response):
    """Test complete workflow from fetch to DataFrame."""
    # Setup mock response
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.text = sample_ecb_response
    mock_handle_api_response.return_value = mock_response

    # Fetch and convert to DataFrame (no url parameter needed)
    df = ecb_instance.to_df()

    # Verify results
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert "time" in df.columns
    assert "currency" in df.columns
    assert "rate" in df.columns

    # Verify we have multiple currencies
    assert df["currency"].nunique() > 1
